import asyncio
import logging
from collections import deque

from nicegui import Event, ui

from .. import rosys
from .communication import Communication
from .esp_pins import EspPins
from .lizard_firmware import LizardFirmware

CLOCK_OFFSET_HISTORY_LENGTH = 100


class RobotBrain:
    """This module manages the communication with a [Zauberzeug Robot Brain](https://zauberzeug.com/products/robot-brain).

    It expects a communication object, which is used for the actual read and write operations.
    Besides providing some basic methods like configuring or restarting the microcontroller, it augments and verifies checksums for each message.

    It also keeps track of the clock offset between the microcontroller and the host system, which is used to synchronize the hardware time with the system time.
    The clock offset is calculated by comparing the hardware time with the system time and averaging the differences over a number of samples.
    If the offset changes significantly, a notification is sent and the offset history is cleared.
    """

    def __init__(self, communication: Communication, *, enable_esp_on_startup: bool = True, use_espresso: bool = False) -> None:
        self.ESP_CONNECTED = Event[[]]()
        """ESP has been connected and Lizard is ready to use"""
        self.LINE_RECEIVED = Event[str]()
        """a line has been received from the microcontroller (argument: line as string)"""
        self.FLASH_P0_COMPLETE = Event[[]]()
        """flashing p0 was successful and 'Replica complete' was received"""

        self.log = logging.getLogger('rosys.robot_brain')

        self.communication = communication
        self.lizard_code = ''
        self.lizard_firmware = LizardFirmware(self)

        self.waiting_list: dict[str, str | None] = {}
        self._clock_offset: float | None = None
        self._clock_offsets: deque[float] = deque(maxlen=CLOCK_OFFSET_HISTORY_LENGTH)
        self._hardware_time: float | None = None
        self._use_espresso = use_espresso
        if enable_esp_on_startup:
            rosys.on_startup(self.enable_esp)

        self.esp_pins_core = EspPins(name='core', robot_brain=self)
        self.esp_pins_p0 = EspPins(name='p0', robot_brain=self)

        self._esp_lock = asyncio.Lock()

    @property
    def clock_offset(self) -> float | None:
        return self._clock_offset

    @property
    def hardware_time(self) -> float | None:
        return self._hardware_time

    @property
    def is_ready(self) -> bool:
        return self._hardware_time is not None

    def developer_ui(self) -> None:
        version_select: ui.select

        async def read_online_versions() -> None:
            await self.lizard_firmware.read_online_version()
            version_select.set_options(list(self.lizard_firmware.online_versions))
            if version_select.options:
                version_select.value = version_select.options[0]

        async def online_update() -> None:
            await self.lizard_firmware.download()
            await self.lizard_firmware.flash_core()
            await self.restart()
            await self.configure()
            await self.lizard_firmware.flash_p0()

        async def local_update() -> None:
            await self.lizard_firmware.flash_core()
            await self.restart()
            await self.configure()
            await self.lizard_firmware.flash_p0()

        with ui.row().classes('items-center'):
            version_select = ui.select([], label='Lizard Version').style('min-width: 140px;') \
                .bind_value_to(self.lizard_firmware, 'selected_online_version')
            ui.button(on_click=read_online_versions).props('icon=refresh flat round dense') \
                .tooltip('Read all available versions from GitHub')
            online_update_button = ui.button(on_click=online_update).props('icon=file_download flat round dense') \
                .tooltip('Download and flash online version to Core and P0 microcontrollers')
        with ui.row().classes('items-center'):
            ui.label().bind_text_from(self.lizard_firmware, 'local_version', backward=lambda x: f'Local: {x or "?"}')
            local_update_button = ui.button(on_click=local_update).props('icon=file_download flat round dense') \
                .tooltip('Flash local version to Core and P0 microcontrollers')
        with ui.row().classes('items-center'):
            ui.label().bind_text_from(self.lizard_firmware, 'core_version', backward=lambda x: f'Core: {x or "?"}')
            configure_button = ui.button(on_click=self.configure).props('icon=build flat round dense') \
                .tooltip('Configure microcontrollers')
        with ui.row().classes('items-center'):
            ui.label().bind_text_from(self.lizard_firmware, 'p0_version', backward=lambda x: f'P0: {x or "?"}')

        def update_visibility() -> None:
            online_update_button.visible = \
                self.lizard_firmware.selected_online_version != self.lizard_firmware.core_version or \
                self.lizard_firmware.selected_online_version != self.lizard_firmware.p0_version
            local_update_button.visible = \
                self.lizard_firmware.local_version != self.lizard_firmware.core_version or \
                self.lizard_firmware.local_version != self.lizard_firmware.p0_version
            configure_button.visible = \
                self.lizard_firmware.local_checksum != self.lizard_firmware.core_checksum
        ui.timer(1.0, update_visibility)

        with ui.row().classes('items-center'):
            ui.button('Refresh', on_click=self.lizard_firmware.read_all).props('outline') \
                .tooltip('Read installed and available versions')
            with ui.row():
                with ui.menu() as menu:
                    ui.menu_item('Download Lizard', on_click=self.lizard_firmware.download) \
                        .tooltip('Download the latest Lizard firmware from GitHub')

                    async def check_strapping_pins_core():
                        ui.notify(await self.esp_pins_core.get_strapping_pins())
                    ui.menu_item('Check Core Strapping Pins', on_click=check_strapping_pins_core) \
                        .tooltip('Check if the Core strapping pins are in the correct state for flashing.')
                    ui.menu_item('Flash Core', on_click=self.lizard_firmware.flash_core) \
                        .tooltip('Flash the downloaded Lizard firmware to the Core microcontroller')

                    async def check_strapping_pins_p0():
                        ui.notify(await self.esp_pins_p0.get_strapping_pins())
                    ui.menu_item('Check P0 Strapping Pins', on_click=check_strapping_pins_p0) \
                        .tooltip('Check if the P0 strapping pins are in the correct state for flashing.')
                    ui.menu_item('Flash P0', on_click=self.lizard_firmware.flash_p0) \
                        .tooltip('Flash the downloaded Lizard firmware to the P0 microcontroller')

                    ui.separator()
                    ui.menu_item('Configure', on_click=self.configure) \
                        .tooltip('Configure the microcontroller with the Lizard startup file')
                    ui.menu_item('Download Config', on_click=lambda: ui.download(self.lizard_code.encode('utf-8'), 'config.liz')) \
                        .tooltip('Download the Lizard config file')

                    ui.separator()
                    ui.menu_item('Enable', on_click=self.enable_esp) \
                        .tooltip('Enable the microcontroller')
                    ui.menu_item('Disable', on_click=self.disable_esp) \
                        .tooltip('Disable the microcontroller')
                    ui.menu_item('Reset', on_click=self.reset_esp) \
                        .tooltip('Reset the microcontroller')
                    ui.menu_item('Restart', on_click=self.restart) \
                        .tooltip('Restart the microcontroller')
                ui.button(on_click=menu.open).props('icon=more_vert flat round')

        ui.label().bind_text_from(self, 'clock_offset', lambda offset: f'Clock offset: {offset or 0:.3f} s')
        ui.label().bind_text_from(self, 'is_ready', lambda ready: f'Ready: {ready}')

    async def configure(self) -> None:
        rosys.notify('Configuring Lizard...')
        await self.send('!-', force=True)
        for line in self.lizard_code.splitlines():
            await self.send(f'!+{line}', force=True)
        await self.send('!.', force=True)
        await self.restart()
        rosys.notify('Lizard configured successfully.', 'positive')

    async def restart(self) -> None:
        await self.send('core.restart()', force=True)
        try:
            await self.LINE_RECEIVED.emitted(timeout=1.0)  # NOTE: we have to wait for the last core message to be sent
        except TimeoutError:
            pass
        finally:
            self._hardware_time = None

    async def read_lines(self) -> list[tuple[float, str]]:
        lines: list[tuple[float, str]] = []
        millis = None
        while True:
            unchecked = await self.communication.read()
            if self._esp_lock.locked():
                break
            line = check(unchecked)
            if not line:
                break
            words = line.split()
            if not words:
                continue
            first = words.pop(0)
            if first in self.waiting_list:
                self.waiting_list[first] = line
            hardware_time: float | None = None
            if first == 'core':
                millis = float(words.pop(0))
                if self.clock_offset is None:
                    continue
                hardware_time = millis / 1000 + self.clock_offset
            if 'Replica complete' in line:
                self.FLASH_P0_COMPLETE.emit()
            self.LINE_RECEIVED.emit(line)
            if hardware_time is not None:
                if self._hardware_time is None:
                    rosys.notify('ESP connected', 'positive')
                    self.ESP_CONNECTED.emit()
                self._hardware_time = hardware_time
            if self._hardware_time is None:
                continue
            lines.append((self._hardware_time, line))
        if millis is not None:
            self._handle_clock_offset(rosys.time() - millis / 1000)
        return lines

    def _handle_clock_offset(self, offset: float) -> None:
        if self._clock_offset is not None and abs(offset - self._clock_offset) > 0.1:
            self.log.info('Clock offset changed from %.3f to %.3f', self._clock_offset, offset)
            self._clock_offsets.clear()
        self._clock_offsets.append(offset)
        self._clock_offset = sum(self._clock_offsets) / len(self._clock_offsets)

    async def send(self, msg: str, *, force: bool = False) -> None:
        """Send a Lizard command to the ESP.

        :param msg: The Lizard command to send
        :param force: Whether to send the message even if the ESP is not ready
        :raises EspNotReadyException: When the ESP is not ready and force is ``False``
        """
        if not self.is_ready and not force:
            raise EspNotReadyException('Sending message failed because ESP is not ready')
        await self.communication.send(augment(msg))

    async def send_and_await(self, msg: str, ack: str, *, timeout: float = float('inf'), force: bool = False) -> str | None:
        """Send a Lizard command to the ESP and await a response.

        :param msg: The Lizard command to send
        :param ack: The first word of the response message to wait for
        :param timeout: Response timeout
        :param force: Whether to send the message even if the ESP is not ready
        :raises EspNotReadyException: When the ESP is not ready and force is ``False``
        :return: The response message or ``None`` if the timeout is reached
        """
        if not self.is_ready and not force:
            raise EspNotReadyException('Sending message failed because ESP is not ready')
        self.waiting_list[ack] = None
        await self.send(msg, force=force)
        t0 = rosys.time()
        while self.waiting_list.get(ack) is None and rosys.time() < t0 + timeout:
            await rosys.sleep(0.1)
        return self.waiting_list.pop(ack) if ack in self.waiting_list else None

    async def enable_esp(self) -> None:
        if self._esp_lock.locked():
            return
        if self._use_espresso:
            await self._enable_espresso()
            return
        self.log.warning("Lizard's flash.py will be deprecated in the future. "
                         'Consider updating Lizard and using the new espresso.py instead.')
        async with self._esp_lock:
            self._hardware_time = None
            rosys.notify('Enabling ESP...')
            command = ['sudo', './flash.py', *self.lizard_firmware.flash_params, 'enable']
            output = await rosys.run.sh(command, timeout=None, working_dir=self.lizard_firmware.PATH)
            self.log.debug(output)
            rosys.notify('Enabling ESP: done')

    async def _enable_espresso(self) -> None:
        if self._esp_lock.locked():
            return
        async with self._esp_lock:
            self._hardware_time = None
            rosys.notify('Enabling ESP...')
            command = ['sudo', './espresso.py', 'enable',
                       *self._convert_flash_params(self.lizard_firmware.flash_params)]
            self.log.debug('enable: %s', command)
            output = await rosys.run.sh(command, timeout=None, working_dir=self.lizard_firmware.PATH)
            if 'Finished.' in output:
                self.log.debug(output)
                rosys.notify('Enabling ESP: done', 'positive')
            else:
                self.log.error(output)
                rosys.notify('Enabling ESP: failed', 'negative')

    async def disable_esp(self) -> None:
        if self._esp_lock.locked():
            return
        async with self._esp_lock:
            self._hardware_time = None
            rosys.notify('Disabling ESP...')
            command = ['sudo', './espresso.py', 'disable', *
                       self._convert_flash_params(self.lizard_firmware.flash_params)]
            self.log.debug('disable: %s', command)
            output = await rosys.run.sh(command, timeout=None, working_dir=self.lizard_firmware.PATH)
            if 'Finished.' in output:
                self.log.debug(output)
                rosys.notify('Disabling ESP: done', 'positive')
            else:
                self.log.error(output)
                rosys.notify('Disabling ESP: failed', 'negative')

    async def reset_esp(self) -> None:
        if self._esp_lock.locked():
            return
        if self._use_espresso:
            await self._reset_espresso()
            return
        self.log.warning("Lizard's flash.py will be deprecated in the future. "
                         'Consider updating Lizard and using the new espresso.py instead.')
        async with self._esp_lock:
            self._hardware_time = None
            rosys.notify('Resetting ESP...')
            command = ['sudo', './flash.py', *self.lizard_firmware.flash_params, 'reset']
            output = await rosys.run.sh(command, timeout=None, working_dir=self.lizard_firmware.PATH)
            self.log.debug(output)
            rosys.notify('Resetting ESP: done')

    async def _reset_espresso(self) -> None:
        if self._esp_lock.locked():
            return
        async with self._esp_lock:
            self._hardware_time = None
            rosys.notify('Resetting ESP...')
            command = ['sudo', './espresso.py', 'reset', *self._convert_flash_params(self.lizard_firmware.flash_params)]
            self.log.debug('reset: %s', command)
            output = await rosys.run.sh(command, timeout=None, working_dir=self.lizard_firmware.PATH)
            if 'Finished.' in output:
                self.log.debug(output)
                rosys.notify('Resetting ESP: done', 'positive')
            else:
                self.log.error(output)
                rosys.notify('Resetting ESP: failed', 'negative')

    def _convert_flash_params(self, flash_params: list[str]) -> list[str]:
        """Until the deprecation of the flash.py script, we need to convert the flash_params to espresso parameters."""
        espresso_parameters = []
        self.log.debug('flash_params: %s', flash_params)
        if 'orin' in flash_params:
            espresso_parameters.extend(['--jetson', 'orin'])
        elif 'xavier' in flash_params:
            espresso_parameters.extend(['--jetson', 'xavier'])
        elif 'nano' in flash_params:
            espresso_parameters.extend(['--jetson', 'nano'])
        if 'nand' in flash_params:
            espresso_parameters.append('--nand')
        if 'orin' in flash_params and 'v05' not in flash_params:
            espresso_parameters.append('--swap_pins')
        self.log.debug('espresso_parameters: %s', espresso_parameters)
        return espresso_parameters

    def __del__(self) -> None:
        self.communication.disconnect()

    def __repr__(self) -> str:
        return f'<RobotBrain {self.communication}>'


class EspNotReadyException(Exception):
    """Raised when trying to use the ESP before it is ready."""


def augment(line: str) -> str:
    checksum = 0
    for c in line:
        checksum ^= ord(c)
    return f'{line}@{checksum:02x}'


def check(line: str | None) -> str:
    if line is None:
        return ''
    if line[-3:-2] != '@':
        return ''
    check_ = int(line[-2:], 16)
    line = line[:-3]
    checksum = 0
    for c in line:
        checksum ^= ord(c)
    if checksum != check_:
        return ''
    return line



from __future__ import annotations

import math

import numpy as np
from nicegui import ui

from ... import rosys
from ...driving.driver import PoseProvider
from ...geometry import GeoPose
from .gnss import Gnss, GnssMeasurement
from .nmea import GpsQuality


class GnssSimulation(Gnss):
    """Simulation of a GNSS receiver."""

    def __init__(self, *,
                 wheels: PoseProvider,
                 lat_std_dev: float = 0.01,
                 lon_std_dev: float = 0.01,
                 heading_std_dev: float = 0.01,
                 interval: float = 1.0,
                 latency: float = 0.0,
                 gps_quality: GpsQuality = GpsQuality.RTK_FIXED) -> None:
        """
        :param wheels: the wheels to use for the simulation
        :param lat_std_dev: the standard deviation of the latitude in meters
        :param lon_std_dev: the standard deviation of the longitude in meters
        :param heading_std_dev: the standard deviation of the heading in degrees
        :param gps_quality: the quality of the GPS signal
        :param interval: the interval between measurements in seconds
        :param latency: the simulated measurement latency in seconds
        """
        super().__init__()
        self.wheels = wheels
        self._is_connected = True
        self._lat_std_dev = lat_std_dev
        self._lon_std_dev = lon_std_dev
        self._heading_std_dev = heading_std_dev
        self._gps_quality = gps_quality
        self._latency = latency
        self.last_measurement: GnssMeasurement | None = None
        rosys.on_repeat(self.simulate, interval)

    @property
    def is_connected(self) -> bool:
        return self._is_connected

    @is_connected.setter
    def is_connected(self, value: bool) -> None:
        self._is_connected = value

    async def simulate(self) -> None:
        if not self.is_connected:
            return
        geo_pose = GeoPose.from_pose(self.wheels.pose)
        noise_lat = np.random.normal(0, self._lat_std_dev)
        noise_lon = np.random.normal(0, self._lon_std_dev)
        noise_magnitude = np.sqrt(noise_lat**2 + noise_lon**2)
        noise_direction = math.atan2(noise_lon, noise_lat)
        noise_point = geo_pose.point.polar(noise_magnitude, noise_direction)
        noise_heading = np.random.normal(geo_pose.heading, math.radians(self._heading_std_dev))
        noise_pose = GeoPose(lat=noise_point.lat, lon=noise_point.lon, heading=noise_heading)
        timestamp = rosys.time()
        self.last_measurement = GnssMeasurement(
            time=timestamp,
            gnss_time=timestamp,
            pose=noise_pose,
            latitude_std_dev=self._lat_std_dev,
            longitude_std_dev=self._lon_std_dev,
            heading_std_dev=self._heading_std_dev,
            gps_quality=self._gps_quality,
        )
        if self._latency:
            await rosys.sleep(self._latency)
        self.NEW_MEASUREMENT.emit(self.last_measurement)

    def developer_ui(self) -> None:
        super().developer_ui()
        with ui.column():
            ui.label('Simulation').classes('text-center text-bold')
            with ui.column().classes('gap-y-1'):
                ui.checkbox('Connected').bind_value(self, '_is_connected')
                ui.number(label='Latitude Std Dev', format='%.3f', prefix='± ', suffix='m') \
                    .bind_value(self, '_lat_std_dev').classes('w-4/5')
                ui.number(label='Longitude Std Dev', format='%.3f', prefix='± ', suffix='m') \
                    .bind_value(self, '_lon_std_dev').classes('w-4/5')
                ui.number(label='Heading Std Dev', format='%.2f', prefix='± ', suffix='°') \
                    .bind_value(self, '_heading_std_dev').classes('w-4/5')
                ui.select({quality: quality.name for quality in GpsQuality}, label='Quality') \
                    .bind_value(self, '_gps_quality').classes('w-4/5')
                ui.number(label='Latency', format='%.2f', suffix='s') \
                    .bind_value(self, '_latency').classes('w-4/5')



import asyncio
import logging
import os
import shlex
import signal
import subprocess
import uuid
from collections.abc import Callable, Coroutine, Generator
from contextlib import contextmanager
from dataclasses import dataclass
from functools import partial, wraps
from inspect import signature
from pathlib import Path
from typing import Any, ParamSpec, TypeVar

from nicegui import run

from . import core
from .helpers import is_stopping

P = ParamSpec('P')
R = TypeVar('R')

running_cpu_bound_processes: list[str] = []  # NOTE is used in tests to advance time slower until computation is done
running_sh_processes: list[subprocess.Popen] = []
log = logging.getLogger('rosys.run')


async def io_bound(callback: Callable[P, R], *args: P.args, **kwargs: P.kwargs) -> R | None:
    if is_stopping():
        return None
    try:
        return await run.io_bound(callback, *args, **kwargs)
    except RuntimeError as e:
        if 'cannot schedule new futures after shutdown' not in str(e):
            raise
    except asyncio.exceptions.CancelledError:
        pass
    return None


def awaitable(func: Callable) -> Callable:
    """decorator to wrap a normal function into an asyncio coroutine"""
    @wraps(func)
    async def inner(*args, **kwargs):
        return await io_bound(func, *args, **kwargs)
    return inner


async def cpu_bound(callback: Callable[P, R], *args: P.args, **kwargs: P.kwargs) -> R | None:
    if is_stopping():
        return None
    with cpu():
        try:
            return await run.cpu_bound(callback, *args, **kwargs)
        except RuntimeError as e:
            if 'cannot schedule new futures after shutdown' not in str(e):
                raise
        except asyncio.exceptions.CancelledError:
            pass
    return None


@contextmanager
def cpu() -> Generator[None, None, None]:
    id_ = str(uuid.uuid4())
    running_cpu_bound_processes.append(id_)
    try:
        yield
    finally:
        running_cpu_bound_processes.remove(id_)


async def sh(command: list[str] | str, *,
             timeout: float | None = 1,
             shell: bool = False,
             working_dir: Path | None = None) -> str:
    """executes a shell command

    Args:
        command: a sequence of program arguments as subprocess.Popen requires or full string
        shell: whether a subshell should be launched (default is `False`, for speed, use `True` if you need file globbing or other features)

    Returns:
        stdout
    """
    def popen() -> str:
        cmd_list = command if isinstance(command, list) else shlex.split(command)
        if timeout is not None:
            cmd_list = ['timeout', '--signal=SIGTERM', str(timeout), *cmd_list]
        cmd = ' '.join(cmd_list) if shell else cmd_list
        try:
            with subprocess.Popen(
                cmd,
                cwd=working_dir,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                shell=shell,
                start_new_session=True,
            ) as proc:
                running_sh_processes.append(proc)
                stdout, stderr = proc.communicate()
                assert proc.stdout is not None
                assert proc.stderr is not None
                proc.stdout.close()
                proc.stderr.close()
                _kill(proc)
                running_sh_processes.remove(proc)
                return stdout.decode('utf-8') if proc.returncode == 0 else stderr.decode('utf-8')
        except Exception:
            log.exception('failed to run command "%s"', cmd)
            if 'proc' in locals() and proc:
                _kill(proc)
            return ''

    if is_stopping():
        return ''
    try:
        async def popen_coro() -> str:
            return await io_bound(popen) or ''
        return await asyncio.wait_for(popen_coro(), timeout)
    except TimeoutError:
        log.warning('Command "%s" timed out after %s seconds.', command, timeout)
        return ''


def _kill(proc: subprocess.Popen) -> None:
    try:
        os.killpg(os.getpgid(proc.pid), signal.SIGTERM)
        log.info('sent SIGTERM to %s', proc.pid)
        try:
            proc.wait(timeout=5)  # wait for 5 seconds
        except subprocess.TimeoutExpired:
            os.killpg(os.getpgid(proc.pid), signal.SIGKILL)  # force kill if process didn't terminate
            log.info('sent SIGKILL to %s', proc.pid)
            proc.wait()  # ensure the process is reaped
    except ProcessLookupError:
        pass
    except Exception:
        log.exception('Failed to kill and/or wait for process %s', proc.pid)


def tear_down() -> None:
    log.info('teardown shell processes...')
    for process in running_sh_processes:
        _kill(process)
    running_sh_processes.clear()
    log.info('teardown complete.')


@dataclass(slots=True, kw_only=True, frozen=True)
class OnFailedArguments:
    attempt: int
    max_attempts: int


async def wait_for(func: Callable[..., Coroutine[Any, Any, R]], timeout: float) -> R:
    """Call a function with a timeout.

    :param func: A function to call
    :param timeout: Maximum time in seconds to wait for the function to return
    :return: Result of the called function
    :raises TimeoutError: If the function does not return within the given time period
    """
    assert timeout > 0, 'timeout must be greater than 0'

    timeout_task = asyncio.create_task(core.sleep(timeout))
    func_task: asyncio.Task[R] = asyncio.create_task(func())
    done, pending = await asyncio.wait([func_task, timeout_task], return_when=asyncio.FIRST_COMPLETED)
    for task in pending:
        task.cancel()
    if timeout_task in done and func_task in pending:
        raise TimeoutError()
    assert func_task in done
    return func_task.result()


async def retry(func: Callable, *,
                max_attempts: int = 3,
                max_timeout: float | None = None,
                on_failed: Callable | None = None) -> Any:
    """Call a function repeatedly until it succeeds or reaches the maximum number of attempts.

    :param func: A function to retry
    :param max_attempts: Maximum number of attempts
    :param max_timeout: Optional maximum time in seconds to wait per attempt
    :param on_failed: Optional callback to execute after each failed attempt (optional argument of type ``OnFailedArguments``)
    :return: Result of the called function
    :raises RuntimeError: If all attempts fail
    """
    for attempt in range(max_attempts):
        try:
            if max_timeout is not None:
                return await wait_for(func, max_timeout)
            else:
                return await func()
        except Exception:
            if on_failed is None:
                continue
            if signature(on_failed).parameters:
                result = on_failed(OnFailedArguments(attempt=attempt, max_attempts=max_attempts))
            else:
                result = on_failed()
            if asyncio.iscoroutinefunction(on_failed):
                await result
    func_name = func.func.__name__ if isinstance(func, partial) else func.__name__
    raise RuntimeError(f'Running {func_name} failed.')



"""Wrapper for TurboJPEG library to normalize return types and provide consistent error handling."""
import numpy as np

try:
    from turbojpeg import TJPF_RGB, TurboJPEG  # type: ignore[import-untyped]
    PY_TURBOJPEG_AVAILABLE = True
except (ModuleNotFoundError, ImportError):
    PY_TURBOJPEG_AVAILABLE = False
    TurboJPEG = None  # type: ignore[assignment, misc]
    TJPF_RGB = 0


class TurboJpegWrapper:
    """Wrapper for TurboJPEG to normalize return types and provide consistent error handling.

    TurboJPEG.encode() can return either bytes or (bytes, int) tuple depending on parameters.
    This wrapper ensures consistent bytes return type and proper exception handling.
    """

    def __init__(self) -> None:
        self._turbo_jpeg = TurboJPEG()  # type: ignore[misc]

    def encode(self, image_array: np.ndarray, quality: int, pixel_format: int = TJPF_RGB) -> bytes:
        """Encode image to JPEG bytes.

        :param image: Image array to encode.
        :param quality: JPEG quality level (1-100).
        :param pixel_format: Pixel format constant from turbojpeg.
        :return: Encoded JPEG bytes.
        :raises RuntimeError: If encoding fails.
        """
        try:
            result = self._turbo_jpeg.encode(image_array, quality=quality, pixel_format=pixel_format)
        except (ValueError, OSError, TypeError) as e:
            raise RuntimeError(f'TurboJPEG encoding failed: {e}') from e

        if isinstance(result, tuple):
            encoded_bytes = result[0]
            if not isinstance(encoded_bytes, bytes):
                raise RuntimeError(f'Unexpected type in tuple from TurboJPEG.encode: {type(encoded_bytes)}')
            return bytes(encoded_bytes)
        if isinstance(result, bytes):
            return result

        raise RuntimeError(f'Unexpected return type from TurboJPEG.encode: {type(result)}')

    def decode(self, jpeg_bytes: bytes, pixel_format: int = TJPF_RGB) -> np.ndarray:
        """Decode JPEG bytes to NumPy array.

        :param jpeg_bytes: JPEG encoded image data.
        :param pixel_format: Pixel format constant from turbojpeg.
        :return: Decoded image as numpy array.
        :raises OSError: If decoding fails.
        """
        try:
            result = self._turbo_jpeg.decode(jpeg_bytes, pixel_format=pixel_format)
        except (ValueError, OSError, TypeError) as e:
            raise OSError(f'TurboJPEG decoding failed: {e}') from e

        if not isinstance(result, np.ndarray):
            raise RuntimeError(f'Unexpected return type from TurboJPEG.decode: {type(result)}')

        return result


TURBO_JPEG: TurboJpegWrapper | None
if PY_TURBOJPEG_AVAILABLE:
    try:
        TURBO_JPEG = TurboJpegWrapper()
    except (RuntimeError, OSError):
        # native library is not available
        TURBO_JPEG = None
else:
    # python wrapper library is not available
    TURBO_JPEG = None



from __future__ import annotations

from collections.abc import Callable, Coroutine
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .rosys import Repeater

# This module gives access to some RoSys functions without needing to import rosys itself.
# This is a workaround to avoid a circular dependency for modules that are part of rosys (e.g. config.py)
# and would need to import rosys (via persistable.py).

is_test: bool
on_repeat: Callable[[Callable, float], Repeater]
on_startup: Callable[[Callable], None]
on_shutdown: Callable[[Callable], None]
sleep: Callable[[float], Coroutine[Any, Any, None]]
time: Callable[[], float]



from __future__ import annotations

import logging
import math
from abc import ABC
from dataclasses import dataclass

from nicegui import Event, ui

from ... import rosys
from ...geometry import GeoPoint, GeoPose
from .nmea import GpsQuality

SECONDS_DAY = 86400
SECONDS_HALF_DAY = 43200


@dataclass(slots=True, kw_only=True)
class GnssMeasurement:
    time: float
    gnss_time: float
    pose: GeoPose
    latitude_std_dev: float = 0.0
    longitude_std_dev: float = 0.0
    heading_std_dev: float = 0.0
    gps_quality: GpsQuality = GpsQuality.INVALID
    num_satellites: int = 0
    hdop: float = 0.0
    altitude: float = 0.0

    @property
    def point(self) -> GeoPoint:
        return self.pose.point

    @property
    def heading(self) -> float:
        return self.pose.heading

    @property
    def age(self) -> float:
        return ((self.gnss_time - rosys.time() + SECONDS_HALF_DAY) % SECONDS_DAY) - SECONDS_HALF_DAY


class Gnss(ABC):
    """A GNSS module that provides measurements from a GNSS receiver."""

    def __init__(self) -> None:
        self.log = logging.getLogger('rosys.gnss')
        self.last_measurement: GnssMeasurement | None = None

        self.NEW_MEASUREMENT = Event[GnssMeasurement]()
        """a new measurement has been received (argument: ``GnssMeasurement``)"""

    @property
    def is_connected(self) -> bool:
        return False

    def developer_ui(self) -> None:
        with ui.column():
            ui.label('GNSS').classes('text-center text-bold')
            with ui.grid(columns='auto auto').classes('gap-y-1 w-2/5'):
                ui.label('Connected:')
                ui.icon('close').bind_name_from(self, 'is_connected',
                                                lambda x: 'check' if x else 'close')
                ui.label('Position:')
                with ui.column().classes('gap-y-0'):
                    ui.label().bind_text_from(self, 'last_measurement',
                                              lambda x: f'{math.degrees(x.pose.lat):.8f}˚' if x else '-')
                    ui.label().bind_text_from(self, 'last_measurement',
                                              lambda x: f'± {x.latitude_std_dev:.3f}m' if x else '-')
                    ui.label().bind_text_from(self, 'last_measurement',
                                              lambda x: f'{math.degrees(x.pose.lon):.8f}˚' if x else '-')
                    ui.label().bind_text_from(self, 'last_measurement',
                                              lambda x: f'± {x.longitude_std_dev:.3f}m' if x else '-')
                ui.label('Heading:')
                with ui.column().classes('gap-y-0'):
                    ui.label().bind_text_from(self, 'last_measurement',
                                              lambda x: f'{math.degrees(x.pose.heading):.2f}˚' if x else '-')
                    ui.label().bind_text_from(self, 'last_measurement',
                                              lambda x: f'± {x.heading_std_dev:.2f}˚' if x else '-')
                ui.label('Quality:')
                ui.label().bind_text_from(self, 'last_measurement',
                                          lambda x: x.gps_quality.name if x else '')
                ui.label('Satellites:')
                ui.label().bind_text_from(self, 'last_measurement',
                                          lambda x: str(x.num_satellites) if x else '-')
                ui.label('HDOP:')
                ui.label().bind_text_from(self, 'last_measurement',
                                          lambda x: f'{x.hdop:.2f}' if x else '-')
                ui.label('Altitude:')
                ui.label().bind_text_from(self, 'last_measurement',
                                          lambda x: f'{x.altitude:.3f}m' if x else '-')
                ui.label('Last update:')
                ui.label().bind_text_from(self, 'last_measurement',
                                          lambda x: f'{rosys.time() - x.time:.2f}s' if x else '')



from dataclasses import dataclass

import cv2
import numpy as np
from scipy.optimize import least_squares

from rosys.geometry import Point3d, Pose3d, Rotation
from rosys.vision.calibration import Calibration, Intrinsics


@dataclass(slots=True, kw_only=True)
class SpatialResectionResult:
    """Result of the spatial resection with lines."""
    success: bool
    iterations: int
    average_reprojection_error: float
    camera_pose: Pose3d
    running_variables: list[float]
    estimated_points_on_lines: list[Point3d]


class SpatialResection:

    def __init__(self, intrinsics: Intrinsics) -> None:
        """Spatial resection to find a camera pose from correspondences between object space and image space."""
        self.intrinsics = intrinsics

    def pnp_with_lsa(self,
                     *,
                     world_points: np.ndarray,
                     image_points: np.ndarray,
                     algorithm: int | str | None = None,
                     p0: Point3d | None = None,
                     r0: Rotation | None = None) -> SpatialResectionResult:
        """Solve the spatial resection problem using OpenCV's PnP and finalize with a least-squares refinement.

        :param world_points: The 3D coordinates of the points in object space, shape (n, 3)
        :param image_points: The 2D coordinates of the points in the image, shape (n, 2)
        :param algorithm: PnP algorithm flag or name (e.g., 'ITERATIVE', 'EPNP', 'P3P', 'AP3P', 'IPPE'). If None, choose automatically.
        :param p0: Initial position of the camera (optional)
        :param r0: Initial rotation of the camera (optional)

        :return: The result of the spatial resection
        """
        assert world_points.shape[0] == image_points.shape[0] and world_points.shape[0] >= 3, 'Need at least 3 correspondences'

        if np.isnan(world_points).any() or np.isnan(image_points).any():
            raise ValueError('Points contain NaNs')

        # Prepare calibration and undistort image observations for a unified pinhole model
        calibration = Calibration(intrinsics=self.intrinsics)
        K_undist = calibration.get_undistorted_camera_matrix()

        object_points = world_points.astype(np.float64).reshape(-1, 1, 3)
        image_points_undist = calibration.undistort_points(image_points.astype(np.float64).reshape(-1, 1, 2))
        D_zeros = np.zeros((1, 5), dtype=np.float64)

        # Decide on algorithm
        def is_planar(points: np.ndarray) -> bool:
            centered = points.reshape(-1, 3) - points.reshape(-1, 3).mean(axis=0)
            _, s, _ = np.linalg.svd(centered, full_matrices=False)
            return s[-1] / max(s[0], 1e-12) < 1e-3

        if algorithm is None:
            # Automatic selection
            num_points = object_points.shape[0]
            if num_points >= 4 and is_planar(object_points):
                method_flag = cv2.SOLVEPNP_IPPE
            elif num_points >= 6:
                method_flag = cv2.SOLVEPNP_EPNP
            elif num_points == 3:
                method_flag = cv2.SOLVEPNP_P3P
            elif num_points >= 4:
                method_flag = cv2.SOLVEPNP_AP3P
            else:
                method_flag = cv2.SOLVEPNP_ITERATIVE
        elif isinstance(algorithm,  int):
            method_flag = algorithm
        elif isinstance(algorithm, str):
            name = algorithm.upper()
            if name == 'ITERATIVE':
                method_flag = cv2.SOLVEPNP_ITERATIVE
            elif name == 'EPNP':
                method_flag = cv2.SOLVEPNP_EPNP
            elif name == 'P3P':
                method_flag = cv2.SOLVEPNP_P3P
            elif name == 'AP3P':
                method_flag = cv2.SOLVEPNP_AP3P
            elif name == 'IPPE':
                method_flag = cv2.SOLVEPNP_IPPE
            else:
                raise ValueError(f'Unknown PnP algorithm: {algorithm}')
        else:
            raise TypeError('Algorithm must be int, str or None')

        # Run PnP on undistorted points with zero distortion and K_undist
        use_guess = p0 is not None and r0 is not None
        if use_guess:
            assert p0 is not None and r0 is not None
            rvec_init = cv2.Rodrigues(r0.T.matrix)[0]
            tvec_init = -r0.T.matrix @ p0.array.reshape(3, 1)
        else:
            rvec_init = None
            tvec_init = None

        ok, rvec, tvec = cv2.solvePnP(
            object_points, image_points_undist, K_undist, D_zeros,
            rvec_init, tvec_init, use_guess, int(method_flag)
        )
        if not ok:
            # Fallback to ITERATIVE
            ok, rvec, tvec = cv2.solvePnP(
                object_points, image_points_undist, K_undist, D_zeros,
                rvec_init, tvec_init, use_guess, int(cv2.SOLVEPNP_ITERATIVE)
            )
        if not ok:
            return SpatialResectionResult(
                success=False,
                iterations=0,
                average_reprojection_error=float('nan'),
                camera_pose=Pose3d(x=float('nan'), y=float('nan'), z=float('nan'), rotation=Rotation.zero()),
                running_variables=[],
                estimated_points_on_lines=[],
            )

        cv2.solvePnPRefineLM(object_points, image_points_undist, K_undist, D_zeros, rvec, tvec)

        Rwc = Rotation.from_rvec(rvec).T
        # Ensure types are numpy arrays so the `@` operator is type-safe for static checkers
        rmat = np.asarray(cv2.Rodrigues(rvec)[0], dtype=np.float64)
        tvec_arr = np.asarray(tvec, dtype=np.float64).reshape(3)
        C = (-rmat.T @ tvec_arr).reshape(3)

        # Compute reprojection error on all observations (undistorted domain)
        proj_all, _ = cv2.projectPoints(object_points, rvec, tvec, K_undist, D_zeros)
        proj_all_np = np.asarray(proj_all, dtype=np.float64).reshape(-1, 2)
        img_all_np = image_points_undist.reshape(-1, 2)
        residuals_all = proj_all_np - img_all_np
        avg_reproj_error = float(np.mean(np.abs(residuals_all)))

        return SpatialResectionResult(
            success=True,
            iterations=1,
            average_reprojection_error=avg_reproj_error,
            camera_pose=Pose3d(x=float(C[0]), y=float(C[1]), z=float(C[2]), rotation=Rwc),
            running_variables=[],
            estimated_points_on_lines=[],
        )

    def lsa_with_lines(self,
                       *,
                       p0: Point3d,
                       r0: Rotation,
                       world_lines: np.ndarray,
                       image_line_points: np.ndarray,
                       s0: np.ndarray | None = None,
                       world_points: np.ndarray | None = None,
                       image_points: np.ndarray | None = None,
                       **ls_args) -> SpatialResectionResult:
        """Solve the spatial resection problem by minimizing the reprojection error.

        The problem is solved using a least squares adjustment to minimize the reprojection error between:

        a) 3D lines and observed 2D points on the lines
        b) 3D points and their 2D projections

        :param p0: The initial position of the camera
        :param r0: The initial rotation of the camera
        :param world_lines: The lines in object space (origin, direction), shape (m, 6)
        :param image_line_points: The 2D coordinates of the points in the image, shape (m, 2)
        :param s0: The initial running variables of the lines, shape (m,) or None
        :param world_points: The 3D coordinates of the points in object space, shape (n, 3) or None
        :param image_points: The 2D coordinates of the points in the image, shape (n, 2) or None
        :param ls_args: Additional arguments for the least squares solver

        :return: The result of the spatial resection
        :raises ValueError: If the Residuals are not finite in the initial state
        """
        # Prepare inputs
        if world_points is None:
            world_points = np.zeros((0, 3))
        if image_points is None:
            image_points = np.zeros((0, 2))
        if s0 is None:
            s0 = np.zeros(len(world_lines))

        # Check sizes
        assert len(world_lines) == len(image_line_points)
        m = len(world_lines)
        assert len(world_points) == len(image_points)
        n = len(world_points)

        # Undistort observations
        calibration = Calibration(intrinsics=self.intrinsics)
        camera_matrix = calibration.get_undistorted_camera_matrix()
        all_image_points = np.concatenate([
            calibration.undistort_points(image_points),
            calibration.undistort_points(image_line_points),
        ])

        # Prepare state vector
        # - 3D position of the camera
        # - 4D quaternion of the camera rotation
        # - m running variables for the lines
        # => [x, y, z, q_w, q_x, q_y, q_z, s_0, s_1, ..., s_m]
        x_0 = np.concatenate([
            p0.array.ravel(),
            r0.T.quaternion,
            s0,
        ])

        # Memory allocation (to avoid reallocations)
        R = np.eye(3).astype(np.float64)
        all_world_points = np.zeros(shape=(n + m, 3))
        image_points_projected = np.zeros(shape=(n + m, 2))
        residuals = np.zeros(shape=(n + m, 2))
        world_line_points = np.zeros(shape=(m, 3))

        def _quaternion_to_rotation_matrix(q: np.ndarray, R: np.ndarray) -> None:
            """Inplace conversion of quaternion to rotation matrix.

            :param q: The quaternion to convert to a rotation matrix: [w, x, y, z]
            :param R: The rotation matrix to fill
            """
            q /= np.linalg.norm(q)
            R[0, 0] = 1 - 2 * q[2]**2 - 2 * q[3]**2
            R[0, 1] = 2 * q[1] * q[2] - 2 * q[0] * q[3]
            R[0, 2] = 2 * q[1] * q[3] + 2 * q[0] * q[2]
            R[1, 0] = 2 * q[1] * q[2] + 2 * q[0] * q[3]
            R[1, 1] = 1 - 2 * q[1]**2 - 2 * q[3]**2
            R[1, 2] = 2 * q[2] * q[3] - 2 * q[0] * q[1]
            R[2, 0] = 2 * q[1] * q[3] - 2 * q[0] * q[2]
            R[2, 1] = 2 * q[2] * q[3] + 2 * q[0] * q[1]
            R[2, 2] = 1 - 2 * q[1]**2 - 2 * q[2]**2

        def f(x: np.ndarray) -> np.ndarray:
            """Functional model using the collinearity equations.

            :param x: The state vector: [x, y, z, q_w, q_x, q_y, q_z, s_0, s_1, ..., s_m]
            :return: The residuals
            """
            # Transform Object coordinates to Camera coordinates
            C = x[:3]
            q = x[3:7]
            _quaternion_to_rotation_matrix(q, R)

            np.multiply(world_lines[:, 3:], x[7:].reshape(-1, 1), out=world_line_points)
            np.add(world_line_points, world_lines[:, :3], out=world_line_points)

            all_world_points[:n] = world_points
            all_world_points[n:] = world_line_points
            np.subtract(all_world_points, C, out=all_world_points)

            # Project 3D points to 2D image points
            Xc = all_world_points @ R.T
            image_points_projected[:] = (Xc @ camera_matrix.T)[:, :2] / Xc[:, 2:3]

            # Compute residuals
            np.subtract(image_points_projected, all_image_points, out=residuals)
            return residuals.flatten()

        res = least_squares(f, x_0, **ls_args)

        return SpatialResectionResult(
            success=res.success,
            iterations=res.nfev,
            average_reprojection_error=np.mean(np.abs(res.fun)),
            camera_pose=Pose3d(x=res.x[0], y=res.x[1], z=res.x[2], rotation=Rotation.from_quaternion(*res.x[3:7]).T),
            running_variables=res.x[7:].tolist(),
            estimated_points_on_lines=[
                Point3d(x=x, y=y, z=z)
                for x, y, z in world_lines[:, :3] + world_lines[:, 3:] * res.x[7:, None]
            ],
        )



from __future__ import annotations

import abc
import json
import logging
import re
import weakref
from pathlib import Path
from typing import Any, ClassVar, Self

from nicegui import run

from .. import core

KEY_PATTERN = re.compile(r'^[a-zA-Z0-9_\-\.]+$')


class Persistable(abc.ABC):
    instances: ClassVar[weakref.WeakValueDictionary[str, Persistable]] = weakref.WeakValueDictionary()

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.log = logging.getLogger('rosys.persistable')
        self._needs_backup = False
        self._filepath: Path | None = None
        self._disabled = False

    def persistent(self, *,
                   key: str | None = None,
                   path: str | Path | None = None,
                   restore: bool = True,
                   backup_check_interval: float | None = 10.0,
                   backup_interval: float | None = None,
                   disable_in_tests: bool = True,
                   ) -> Self:
        """Make this object persistent.

        If ``restore`` is ``True``, the object will be restored immediately from the persistence file.
        If ``backup_check_interval`` is not ``None`` or 0.0, the object will be backed up at the given interval _if a backup has been requested_.
        If ``backup_interval`` is not ``None`` or 0.0, the object will be backed up at the given interval _independently of any backup requests_.

        :param key: The key to use for the persistence file (letters, digits, dashes and underscores; default: module name)
        :param path: The path to the persistence file (default: "~/.rosys")
        :param restore: Whether to restore the object immediately from the persistence file (default: True)
        :param backup_check_interval: The interval to check for backups (default: 10.0 seconds)
        :param backup_interval: The interval to backup the object (default: None)
        :param disable_in_tests: Whether to disable persistence in tests (default: True)
        """
        self._disabled = core.is_test and disable_in_tests
        if self._disabled:
            return self
        if self._filepath is not None:
            raise RuntimeError('This object is already persistent')

        if key is None:
            key = self.__module__
        if not key.strip('.'):
            raise ValueError('Key is empty')
        if not KEY_PATTERN.match(key):
            raise ValueError(f'Key "{key}" contains invalid characters')
        if key in self.instances:
            raise ValueError(f'Key "{key}" is already used by another persistent object')
        self.instances[key] = self

        directory = Path(path or '~/.rosys').expanduser()
        directory.mkdir(parents=True, exist_ok=True)
        self._filepath = directory / f'{key}.json'
        if restore:
            self.sync_restore()
        if backup_check_interval:
            core.on_repeat(self._backup_if_requested, backup_check_interval)
        if backup_interval:
            core.on_repeat(self.backup, backup_interval)
        if backup_interval:
            core.on_shutdown(self.sync_backup)
        elif backup_check_interval:
            core.on_shutdown(self._sync_backup_if_requested)
        return self

    def request_backup(self) -> None:
        """Request a backup of the object.

        This will trigger a backup at the check defined by ``backup_check_interval``.
        """
        self._needs_backup = True

    async def _backup_if_requested(self) -> None:
        if self._needs_backup:
            await self.backup()

    def _sync_backup_if_requested(self) -> None:
        if self._needs_backup:
            self.sync_backup()

    async def backup(self) -> None:
        """Backup the object.

        This will immediately write the object to the persistence file.
        The file is written in an I/O bound task to avoid blocking the main thread.
        """
        await run.io_bound(self.sync_backup)

    def sync_backup(self) -> None:
        """Backup the object.

        This will immediately write the object to the persistence file.
        The file is written synchronously.
        To avoid blocking the main thread, use ``request_backup()`` instead.
        """
        if self._disabled:
            raise RuntimeError('Backup failed: Persistence is disabled.')
        if self._filepath is None:
            raise RuntimeError('Backup failed: This object is not persistent. Call persistent() first.')
        self._filepath.write_text(json.dumps(self.backup_to_dict()))
        self._needs_backup = False

    async def restore(self) -> None:
        """Restore the object from the persistence file.

        The file is read in an I/O bound task to avoid blocking the main thread.
        """
        await run.io_bound(self.sync_restore)

    def sync_restore(self) -> None:
        """Restore the object from the persistence file.

        The file is read synchronously.
        To avoid blocking the main thread, use ``restore()`` instead.
        """
        if self._disabled:
            raise RuntimeError('Restore failed: Persistence is disabled.')
        if self._filepath is None:
            raise RuntimeError('Restore failed: This object is not persistent. Call persistent() first.')
        if not self._filepath.exists():
            self.log.warning('Restore failed: File "%s" not found', self._filepath)
            return
        self.restore_from_dict(json.loads(self._filepath.read_text()))

    @abc.abstractmethod
    def backup_to_dict(self) -> dict[str, Any]:
        """Convert the object to a dictionary."""

    @abc.abstractmethod
    def restore_from_dict(self, data: dict[str, Any]) -> None:
        """Restore the object from a dictionary."""



import logging
from datetime import datetime
from pathlib import Path

from nicegui import ui

from ... import rosys
from ..camera_provider import CameraProvider
from .replay_camera import ReplayCamera

log = logging.getLogger('rosys.replay_camera_provider')


class ReplayCameraProvider(CameraProvider[ReplayCamera]):
    """This module collects and simulates cameras by looping over images in a drive.

    NOTE: The first image of each camera will be emitted at any time before the timestamp of the second image.
    Therefore, setting the time to 0.0 will cause all cameras to emit their first image.

    :param replay_folder: path to the root replay folder containing the camera folders
    :param replay_interval: the interval at which the provider checks for new images (in seconds)
    """

    def __init__(self, replay_folder: Path, replay_interval: float = .01) -> None:
        super().__init__()

        self._start_time: float = float('inf')
        self._current_time: float = 0.0
        self._cameras_need_update = False
        self._end_time: float = 0.0

        self._playback_speed = 1.0

        self._running = True

        self._last_update_time: float = rosys.time()

        self._find_cameras(replay_folder)
        self._set_time_interval()

        self._repeater = rosys.on_repeat(self._step, interval=replay_interval)

    def pause(self) -> None:
        self._update_time()
        self._running = False

    def play(self) -> None:
        self._last_update_time = rosys.time()
        self._running = True

    def stop(self) -> None:
        self._running = False
        self.jump_to(0.0)

    def toggle_running(self) -> None:
        if self._running:
            self.pause()
        else:
            self.play()

    def create_ui(self, *, skip_time: float = 2.0, time_label_format: str = '%d.%m.%Y %H:%M:%S') -> None:
        """Create a simple UI for controlling the replay.

        :param skip_time: the amount of time to skip when pressing the skip buttons (in seconds)
        :param time_label_format: the format to be used for the current time label (used by ``datetime.strftime``)
        """
        with ui.row(align_items='center').classes('w-full gap-2'):
            ui.button(icon='sym_o_replay') \
                .on_click(lambda: self.jump_to_time(max(self._current_time - skip_time, self._start_time))) \
                .props('dense size=md') \
                .tooltip(f'<- {skip_time}s')
            ui.button(icon='play_arrow', on_click=self.toggle_running) \
                .bind_icon_from(self, '_running', backward=lambda p: 'pause' if p else 'play_arrow') \
                .props('dense size=md')
            ui.button(icon='sym_o_forward_media') \
                .on_click(lambda: self.jump_to_time(min(self._current_time + skip_time, self._end_time))) \
                .props('dense size=md') \
                .tooltip(f'-> {skip_time}s')
            ui.slider(min=0.25, max=4, step=0.25, on_change=lambda e: self.set_speed(e.value)) \
                .bind_value_from(self, '_playback_speed') \
                .props('label snap') \
                .classes('w-32 ml-4')
        s = ui.slider(min=self._start_time, max=self._end_time) \
            .bind_value(self, '_current_time') \
            .on('change', lambda e: self.jump_to_time(e.args)) \
            .props('label-always') \
            .classes('w-full')
        s.bind_value_to(s.props, 'label-value', lambda x: f'{datetime.fromtimestamp(x).strftime(time_label_format)}')

    def set_speed(self, speed: float) -> None:
        """Set the playback speed.

        1.0 is real-time, 2.0 is double speed, 0.5 is half speed.
        You may use negative values to play backwards.
        """
        self._playback_speed = speed

    def jump_to(self, percent: float) -> None:
        """Jump to a specific point in the replay.

        :param percent: a value between 0.0 and 100.0, where 0.0 is the start and 100.0 is the end
        """
        assert 0.0 <= percent <= 100.0, 'Percent must be between 0.0 and 100.0'
        self.jump_to_time(self._start_time + (percent / 100.0) * (self._end_time - self._start_time))

    def jump_to_time(self, time: float) -> None:
        """Jump to a specific time in the replay.

        :param time: time in seconds
        """
        assert self._start_time <= time <= self._end_time, f'Time must be between {self._start_time} & {self._end_time}'
        self.clear_camera_images()
        self._set_replay_time(time)

    def _set_replay_time(self, time: float) -> None:
        self._current_time = time
        self._cameras_need_update = True

    def _find_cameras(self, replay_folder: Path) -> None:
        for file in replay_folder.iterdir():
            if file.is_dir():
                cam_name = file.name.replace('--', ':')
                camera = ReplayCamera(camera_id=cam_name, images_dir=file, camera_name=cam_name)
                self.add_camera(camera)
                self.log.info('Added replay camera "%s" from folder "%s"', cam_name, file)

    def _set_time_interval(self) -> None:
        for camera in self.cameras.values():
            if not camera.image_paths:
                continue

            timestamps = camera.timestamp_array
            self._start_time = min(self._start_time, timestamps[0])
            self._end_time = max(self._end_time, timestamps[-1])
            self._current_time = self._start_time

    async def _step(self) -> None:
        self._update_time()
        await self._update_camera_images()

    def _update_time(self) -> None:
        if not self._running:
            return

        now = rosys.time()
        new_replay_time = self._current_time + (now - self._last_update_time) * self._playback_speed
        if new_replay_time > self._end_time:
            new_replay_time = self._start_time + (new_replay_time - self._end_time)
        self._set_replay_time(new_replay_time)
        self._last_update_time = now

    async def _update_camera_images(self) -> None:
        if not self._cameras_need_update:
            return
        for camera in self.cameras.values():
            await camera.load_image_at_time(self._current_time)

    def clear_camera_images(self) -> None:
        for camera in self.cameras.values():
            camera.images.clear()

    async def update_device_list(self) -> None:
        pass



from __future__ import annotations

import logging
from dataclasses import dataclass, field
from enum import Enum
from typing import cast, overload

import cv2
import numpy as np
from numpy.typing import NDArray

from ..geometry import Frame3d, Point, Point3d, Pose3d, Rotation
from .image import Image, ImageSize

FloatArray = NDArray[np.float32] | NDArray[np.float64]


class CameraModel(str, Enum):
    PINHOLE = 'pinhole'
    FISHEYE = 'fisheye'
    OMNIDIRECTIONAL = 'omnidirectional'


@dataclass(slots=True, kw_only=True)
class OmnidirParameters:
    """The parameters for an omnidirectional camera.

    :param xi: The omnidirectional camera parameter xi.
    :param rotation: Rotation used to determine crop when undistorting to pinhole image.
    """
    xi: float = 0.0
    rotation: Rotation = field(default_factory=Rotation.zero)


@dataclass(slots=True, kw_only=True)
class Intrinsics:
    """The intrinsic parameters of a camera.

    :param model: The camera model to use.
    :param matrix: The camera matrix K.
    :param distortion: The distortion coefficients D.
    :param xi: The omnidirectional camera parameter xi (only for ``CameraModel.OMNIDIRECTIONAL``).
    :param rotation: An inner rotation matrix, useful for visual-inertial calibration or omnidirectional projection.
    :param size: The size of the image.
    """
    model: CameraModel = CameraModel.PINHOLE
    matrix: list[list[float]]
    distortion: list[float]
    omnidir_params: OmnidirParameters | None = None
    size: ImageSize

    @staticmethod
    def create_default(width: int = 800,
                       height: int = 600, *,
                       focal_length: float = 570,
                       distortion: list[float] | None = None,
                       ) -> Intrinsics:
        return Intrinsics(
            matrix=[
                [focal_length, 0, width / 2],
                [0, focal_length, height / 2],
                [0, 0, 1],
            ],
            distortion=distortion or [0.0, 0.0, 0.0, 0.0, 0.0],
            size=ImageSize(width=width, height=height),
        )


log = logging.getLogger('rosys.vision.calibration')


@dataclass(slots=True, kw_only=True)
class Calibration:
    """Represents the full calibration of a camera."""
    intrinsics: Intrinsics
    extrinsics: Pose3d = field(default_factory=Pose3d)

    @staticmethod
    def from_points(world_points: list[Point3d] | list[list[Point3d]],
                    image_points: list[Point] | list[list[Point]],
                    *,
                    image_size: ImageSize,
                    f0: float,
                    rational_model: bool = False,
                    camera_model: CameraModel = CameraModel.PINHOLE,
                    frame: Frame3d | None = None) -> Calibration:
        """Estimate the camera calibration from corresponding world and image points.

        :param world_points: The observed points in 3D world coordinates.
        :param image_points: The observed points in 2D image coordinates.
        :param image_size: The size of the image.
        :param f0: An initial guess for the focal length.
        :param rational_model: Whether to use the rational camera model (only applies to pinhole cameras).
        :param camera_model: The camera model to use.
        :param frame: The coordinate frame of the world points and the extrinsic camera pose.

        :return: The estimated camera calibration.
        """
        if rational_model and camera_model != CameraModel.PINHOLE:
            raise ValueError('Rational model is only supported for pinhole cameras')

        if isinstance(world_points[0], Point3d) or isinstance(image_points[0], Point):
            assert len(world_points) == len(image_points), 'Image and world points are not formatted equally'
            world_points = [world_points]  # type: ignore
            image_points = [image_points]  # type: ignore

        world_point_array = [np.array([p.tuple for p in view], dtype=np.float32).reshape((-1, 1, 3))  # type: ignore
                             for view in world_points]
        image_point_array = [np.array([p.tuple for p in view], dtype=np.float32).reshape((-1, 1, 2))  # type: ignore
                             for view in image_points]
        K0 = np.array([[f0, 0, image_size.width / 2], [0, f0, image_size.height / 2], [0, 0, 1]], dtype=np.float32)
        D0 = np.array([0.1, 0.4, -0.5, 0.2], dtype=np.float32).reshape(1, 4)

        xi: float = 0.0

        optimization_criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 200, 1e-6)

        if camera_model == CameraModel.PINHOLE:
            flags = cv2.CALIB_USE_INTRINSIC_GUESS
            if rational_model:
                flags |= cv2.CALIB_RATIONAL_MODEL
            _rms, K, D, rvecs, tvecs = cv2.calibrateCamera(
                objectPoints=world_point_array,
                imagePoints=image_point_array,
                imageSize=image_size.tuple,
                cameraMatrix=K0,
                distCoeffs=np.zeros((1, 5), dtype=np.float32),
                flags=flags,
                criteria=optimization_criteria,
            )
        elif camera_model == CameraModel.FISHEYE:
            flags = cv2.fisheye.CALIB_RECOMPUTE_EXTRINSIC
            flags |= cv2.fisheye.CALIB_CHECK_COND
            flags |= cv2.fisheye.CALIB_FIX_SKEW
            flags |= cv2.fisheye.CALIB_USE_INTRINSIC_GUESS
            _rms, K, D, rvecs, tvecs = cv2.fisheye.calibrate(  # pylint: disable=unpacking-non-sequence
                objectPoints=world_point_array,
                imagePoints=image_point_array,
                image_size=image_size.tuple,
                K=K0,
                D=D0,
                flags=flags,
                criteria=optimization_criteria,
            )
            if D.size != 4:
                raise ValueError(f'Fisheye calibration failed (got invalid distortion coefficients D="{D}")')
        elif camera_model == CameraModel.OMNIDIRECTIONAL:
            flags = cv2.omnidir.CALIB_USE_GUESS
            flags |= cv2.omnidir.CALIB_FIX_SKEW
            _rms, K, xi_arr, D, rvecs, tvecs, _status = cv2.omnidir.calibrate(  # pylint: disable=unpacking-non-sequence
                objectPoints=world_point_array,
                imagePoints=image_point_array,
                size=image_size.tuple,
                K=K0,
                xi=np.array([xi], dtype=np.float32),
                D=D0,
                flags=flags,
                criteria=optimization_criteria,
            )
            xi = xi_arr[0]
            if D.size != 4:
                raise ValueError(f'Omnidirectional calibration failed (got invalid distortion coefficients D="{D}")')
        else:
            raise ValueError(f'Unknown camera model "{camera_model}"')

        intrinsics = Intrinsics(matrix=K.tolist(),
                                distortion=D.tolist()[0],
                                size=image_size,
                                model=camera_model,
                                omnidir_params=OmnidirParameters(xi=xi))

        rotation = Rotation.from_rvec(rvecs[0]).T
        translation = -np.array(rotation.R).dot(tvecs[0])
        extrinsics = Pose3d(x=translation[0, 0],
                            y=translation[1, 0],
                            z=translation[2, 0],
                            rotation=rotation).in_frame(frame)

        return Calibration(intrinsics=intrinsics, extrinsics=extrinsics)

    @overload
    def project_to_image(self, coordinates: Point3d) -> Point | None:
        """Project a 3D point to the image plane.

        :param coordinates: The 3D point to project.
        :return: The image coordinates of the projected point, or None if projection failed.
        """

    @overload
    def project_to_image(self, coordinates: list[Point3d]) -> list[Point | None]:
        """Project 3D points to the image plane.

        :param coordinates: The list of 3D points to project.
        :return: The image coordinates of the projected points, or None if any projection failed.
        """

    @overload
    def project_to_image(self, coordinates: FloatArray, *, frame: Frame3d | None = None) -> FloatArray:
        """Project 3D points to the image plane.

        :param coordinates: (Nx3) The 3D coordinates of the points to project.
        :param frame: The coordinate frame of the coordinates.
        :return: (Nx2) The image coordinates of the projected points.
        """

    def project_to_image(self, coordinates: Point3d | list[Point3d] | FloatArray, *, frame: Frame3d | None = None) -> Point | None | list[Point | None] | FloatArray:

        if isinstance(coordinates, Point3d):
            return self.project_to_image([coordinates])[0]
        if isinstance(coordinates, list):
            world_array = np.array([point.resolve().tuple for point in coordinates], dtype=np.float64)
            image_array = self.project_to_image(world_array)
            return [Point(x=vector[0], y=vector[1]) if not np.isnan(vector).any() else None for vector in image_array]

        if len(coordinates) == 0:
            return np.empty((0, 2), dtype=np.float64)

        world_extrinsics = self.extrinsics.relative_to(frame)
        R = world_extrinsics.rotation.matrix.astype(np.float64)
        Rod = cv2.Rodrigues(R.T)[0]
        t = -R.T @ world_extrinsics.translation
        K = np.array(self.intrinsics.matrix)
        D = np.array(self.intrinsics.distortion, dtype=float)

        # pylint: disable=unpacking-non-sequence
        if self.intrinsics.model == CameraModel.PINHOLE:
            image_array, _ = cast(FloatArray, cv2.projectPoints(coordinates, Rod, t, K, D))
        elif self.intrinsics.model == CameraModel.FISHEYE:
            image_array, _ = cast(FloatArray, cv2.fisheye.projectPoints(coordinates.reshape(-1, 1, 3), Rod, t, K, D))
        elif self.intrinsics.model == CameraModel.OMNIDIRECTIONAL:
            assert self.intrinsics.omnidir_params is not None, 'Omnidirectional parameters are unset'
            xi = self.intrinsics.omnidir_params.xi
            image_array, _ = cast(FloatArray, cv2.omnidir.projectPoints(
                coordinates.reshape(-1, 1, 3), Rod, t, K, xi, D))
        else:
            raise ValueError(f'Unknown camera model "{self.intrinsics.model}"')
        # pylint: enable=unpacking-non-sequence

        if self.intrinsics.model != CameraModel.OMNIDIRECTIONAL:
            coordinates = coordinates.reshape(-1, 3)
            local_coordinates = (coordinates - world_extrinsics.translation) @ R
            image_array[local_coordinates[:, 2] < 0, :] = np.nan

        return image_array.reshape(-1, 2)

    @overload
    def project_from_image(self,
                           image_coordinates: Point,
                           target_height: float = 0, *,
                           reprojection_tolerance: float = 0.002) -> Point3d | None:
        """Project a point in image coordinates to a plane in world coordinates.

        :param image_coordinates: The image coordinates to project.
        :param target_height: The height of the plane in world coordinates.
        :param reprojection_tolerance: The reprojection tolerance as a fraction of the maximum image dimension.
        :return: The world coordinates of the projected point, or None if projection failed.
        """

    @overload
    def project_from_image(self,
                           image_coordinates: list[Point],
                           target_height: float = 0, *,
                           reprojection_tolerance: float = 0.002) -> list[Point3d | None]:
        """Project points in image coordinates to a plane in world coordinates.

        :param image_coordinates: The list of image coordinates to project.
        :param target_height: The height of the plane in world coordinates.
        :param reprojection_tolerance: The reprojection tolerance as a fraction of the maximum image dimension.
        :return: The world coordinates of the projected points, or None if any projection failed.
        """

    @overload
    def project_from_image(self,
                           image_coordinates: FloatArray,
                           target_height: float = 0, *,
                           reprojection_tolerance: float = 0.002) -> FloatArray:
        """Project points in image coordinates to a plane in world coordinates.

        :param image_coordinates: (Nx2) The image coordinates to project.
        :param target_height: The height of the plane in world coordinates.
        :param reprojection_tolerance: The reprojection tolerance as a fraction of the maximum image dimension.
        :return: (Nx3) The world coordinates of the projected points.
        """

    def project_from_image(self,
                           image_coordinates: Point | list[Point] | FloatArray,
                           target_height: float = 0, *,
                           reprojection_tolerance: float = 0.002) -> Point3d | None | list[Point3d | None] | FloatArray:
        if isinstance(image_coordinates, Point):
            return self.project_from_image([image_coordinates],
                                           target_height=target_height,
                                           reprojection_tolerance=reprojection_tolerance)[0]
        if not isinstance(image_coordinates, np.ndarray):
            image_array = np.array([point.tuple for point in image_coordinates], dtype=np.float64)
            world_array = self.project_from_image(image_array,
                                                  target_height=target_height,
                                                  reprojection_tolerance=reprojection_tolerance)
            return [Point3d(x=point[0], y=point[1], z=point[2]) if not np.isnan(point).any() else None for point in world_array]

        if len(image_coordinates) == 0:
            return np.empty((0, 3), dtype=np.float64)

        world_extrinsics = self.extrinsics.resolve()
        image_rays = self._points_to_rays(image_coordinates.astype(np.float64).reshape(-1, 1, 2))
        objPoints = image_rays @ world_extrinsics.rotation.matrix.T
        Z = world_extrinsics.z
        t = world_extrinsics.translation_vector
        world_array = t.T - objPoints * (Z - target_height) / objPoints[:, 2:]

        # Check if the point is in front of the camera
        sign = objPoints[:, -1] * np.sign(Z)
        world_array[sign >= 0, :] = np.nan

        # Check if the point is within the reprojection tolerance
        if reprojection_tolerance != float('inf'):
            tolerance_px = max(self.intrinsics.size.width, self.intrinsics.size.height) * reprojection_tolerance
            reprojection = self.project_to_image(world_array)
            distance = np.linalg.norm(reprojection - image_coordinates.reshape(-1, 2), axis=1)
            world_array[distance > tolerance_px, :] = np.nan

        return world_array

    def _points_to_rays(self, image_points: np.ndarray) -> np.ndarray:
        """Convert image points to rays in homogeneous coordinates with respect to the camera coordinate frame."""
        K = np.array(self.intrinsics.matrix, dtype=np.float32).reshape((3, 3))
        D = np.array(self.intrinsics.distortion)
        if self.intrinsics.model == CameraModel.PINHOLE:
            undistorted = cv2.undistortPoints(image_points, K, D)
        elif self.intrinsics.model == CameraModel.FISHEYE:
            undistorted = cv2.fisheye.undistortPoints(image_points, K, D)
        elif self.intrinsics.model == CameraModel.OMNIDIRECTIONAL:
            assert self.intrinsics.omnidir_params is not None, 'Omnidirectional parameters are unset'
            R = self.intrinsics.omnidir_params.rotation.matrix.astype(np.float32)
            xi = np.array(self.intrinsics.omnidir_params.xi, dtype=np.float32)
            undistorted = cv2.omnidir.undistortPoints(image_points, K, D, xi=xi, R=R)
        else:
            raise ValueError(f'Unknown camera model "{self.intrinsics.model}"')

        return cv2.convertPointsToHomogeneous(undistorted).reshape(-1, 3)

    @overload
    def undistort_points(self, image_points: Point, *, crop: bool = False) -> Point:
        """Undistort an image point using the camera calibration.

        :param image_points: The image point to undistort.
        :param crop: Whether cropping is applied to the image during undistortion.
        :return: The undistorted image point.
        """

    @overload
    def undistort_points(self, image_points: list[Point], *, crop: bool = False) -> list[Point]:
        """Undistort image points using the camera calibration.

        :param image_points: The list of image points to undistort.
        :param crop: Whether cropping is applied to the image during undistortion.
        :return: The undistorted image points as a list of Points.
        """

    @overload
    def undistort_points(self, image_points: FloatArray, *, crop: bool = False) -> FloatArray:
        """Undistort image points using the camera calibration.

        :param image_points: (Nx2) The image points to undistort.
        :param crop: Whether cropping is applied to the image during undistortion.
        :return: (Nx2) The undistorted image points.
        """

    def undistort_points(self, image_points: Point | list[Point] | FloatArray, *, crop: bool = False) -> Point | list[Point] | FloatArray:
        if isinstance(image_points, Point):
            return self.undistort_points([image_points], crop=crop)[0]

        if len(image_points) == 0:
            return image_points

        if not isinstance(image_points, np.ndarray):
            image_points = np.array([p.tuple for p in image_points], dtype=np.float32)
            image_points = self.undistort_points(image_points, crop=crop)
            return [Point(x=p[0], y=p[1]) for p in image_points]

        image_points = image_points.reshape(-1, 1, 2)

        K = np.array(self.intrinsics.matrix, dtype=np.float32).reshape((3, 3))
        D = np.array(self.intrinsics.distortion)

        if self.intrinsics.model == CameraModel.PINHOLE:
            if crop:
                log.warning('Cropping is not yet supported for pinhole cameras')
            new_K = self.get_undistorted_camera_matrix(crop=False)
            return cast(FloatArray, cv2.undistortPoints(image_points, K, D, P=new_K, R=np.eye(3)).reshape(-1, 2))
        elif self.intrinsics.model == CameraModel.FISHEYE:
            new_K = self.get_undistorted_camera_matrix(crop=crop)
            return cast(FloatArray, cv2.fisheye.undistortPoints(image_points, K, D, P=new_K).reshape(-1, 2))
        elif self.intrinsics.model == CameraModel.OMNIDIRECTIONAL:
            assert self.intrinsics.omnidir_params is not None, 'Omnidirectional parameters are unset'
            R = np.array(self.intrinsics.omnidir_params.rotation, dtype=np.float32)
            xi = np.array(self.intrinsics.omnidir_params.xi, dtype=np.float32)
            return cast(FloatArray, cv2.omnidir.undistortPoints(image_points, K, D, xi=xi, R=R).reshape(-1, 2))
        else:
            raise ValueError(f'Unknown camera model "{self.intrinsics.model}"')

    FloatArray = NDArray[np.float32] | NDArray[np.float64]

    @overload
    def distort_points(self, image_points: Point, *, crop: bool = False) -> Point:
        """Distort an image point using the camera calibration.

        :param image_points: The image point to distort.
        :param crop: Whether cropping is applied to the image during distortion.
        :return: The distorted image point.
        """

    @overload
    def distort_points(self, image_points: list[Point], *, crop: bool = False) -> list[Point]:
        """Distort a list of image points.
        Note: For pinhole models the redistortion can be off by more than 1px for large distortions.

        :param image_points: The list of image points to distort.
        :param crop: Whether cropping is applied to the image during distortion.
        :return: The distorted image points as a list of Points.
        """

    @overload
    def distort_points(self, image_points: FloatArray, *, crop: bool = False) -> FloatArray:
        """Apply lens distortion to image points using the camera calibration.
        Note: For pinhole models the redistortion can be off by more than 1px for large distortions.

        :param image_points: (Nx2) The image points to distort.
        :param crop: Whether cropping is applied to the image during distortion.
        :return: (Nx2) The distorted image points.
        """

    def distort_points(self, image_points: Point | list[Point] | FloatArray, *, crop: bool = False) -> Point | list[Point] | FloatArray:
        if isinstance(image_points, Point):
            return self.distort_points([image_points], crop=crop)[0]

        if len(image_points) == 0:
            return image_points

        if not isinstance(image_points, np.ndarray):
            image_points = np.array([p.tuple for p in image_points], dtype=np.float32)
            image_points = self.distort_points(image_points, crop=crop)
            return [Point(x=p[0], y=p[1]) for p in image_points]

        if image_points.dtype not in [np.float32, np.float64]:
            raise ValueError('Image point array must be of type float32 or float64')

        image_points = image_points.reshape(-1, 1, 2)

        K = np.array(self.intrinsics.matrix, dtype=np.float32).reshape((3, 3))
        D = np.array(self.intrinsics.distortion)

        if self.intrinsics.model == CameraModel.PINHOLE:
            return self._distort_points_pinhole(image_points).reshape(-1, 2)
        elif self.intrinsics.model == CameraModel.FISHEYE:
            new_K = self.get_undistorted_camera_matrix(crop=crop)
            normalized_points = cv2.undistortPoints(image_points, new_K, None)  # type: ignore
            return cast(FloatArray, cv2.fisheye.distortPoints(normalized_points, K, D).reshape(-1, 2))
        elif self.intrinsics.model == CameraModel.OMNIDIRECTIONAL:
            raise NotImplementedError('Re-distortion for omnidirectional cameras is not supported')
        else:
            raise ValueError(f'Unknown camera model "{self.intrinsics.model}"')

    def _distort_points_pinhole(self, image_points: np.ndarray) -> np.ndarray:
        K = np.array(self.intrinsics.matrix, dtype=np.float32)
        D = np.array(self.intrinsics.distortion, dtype=np.float32)
        new_K = self.get_undistorted_camera_matrix(crop=False)

        # Note: this is a slight hack of available functions.
        # We first normalize the points, then project them
        # In the end this applies only the distortion during projectPoints
        normalized_points = cv2.undistortPoints(image_points, new_K, None)  # type: ignore
        points_3d = cv2.convertPointsToHomogeneous(normalized_points)
        distorted_points, _ = cv2.projectPoints(points_3d, np.zeros(3), np.zeros(3), K, D)
        return distorted_points

    def get_undistorted_camera_matrix(self, crop: bool = False) -> np.ndarray:
        """Compute the camera matrix for the undistorted image.

        :param crop: Whether cropping is applied to the image during undistortion.

        :return: The camera matrix for the undistorted image.
        """
        K = np.array(self.intrinsics.matrix)
        D = np.array(self.intrinsics.distortion)
        w, h = self.intrinsics.size.width, self.intrinsics.size.height
        if self.intrinsics.model == CameraModel.PINHOLE:
            new_K, _ = cv2.getOptimalNewCameraMatrix(K, D, (w, h), 1, (w, h), centerPrincipalPoint=True)
        elif self.intrinsics.model == CameraModel.FISHEYE:
            new_K = cv2.fisheye.estimateNewCameraMatrixForUndistortRectify(K, D, (w, h), np.eye(3),
                                                                           balance=0.5 if crop else 1.0,
                                                                           new_size=(w, h),
                                                                           fov_scale=1)
        elif self.intrinsics.model == CameraModel.OMNIDIRECTIONAL:
            if crop:
                log.warning('Cropping is not yet supported for omnidirectional cameras')
            new_K = np.array([[w / 4, 0, w / 2],
                              [0, h / 4, h / 2],
                              [0, 0, 1]])
        else:
            raise ValueError(f'Unknown camera model "{self.intrinsics.model}"')

        return new_K

    @overload
    def undistort_image(self, image: Image, *, crop: bool = False) -> Image:
        """Undistort an image represented as an Image object.

        If you already have the image as an unencoded numpy array, use ``undistort_array`` instead.

        :param image: The image to undistort.
        :param crop: Whether cropping is applied to the image during undistortion.

        :return: The undistorted image.
        """

    @overload
    def undistort_image(self, image: NDArray, *, crop: bool = False) -> NDArray:
        """Undistort an image represented as a numpy array.

        :param image: The image to undistort.
        :param crop: Whether cropping is applied to the image during undistortion.

        :return: The undistorted image.
        """

    def undistort_image(self, image: Image | NDArray, *, crop: bool = False) -> Image | NDArray:

        if isinstance(image, Image):
            image_array = self.undistort_image(image.array, crop=crop)
            return Image.from_array(image_array, camera_id=image.camera_id, time=image.time, metadata=image.metadata)

        if image.shape[0] != self.intrinsics.size.height or image.shape[1] != self.intrinsics.size.width:
            log.warning('Image size does not match calibration size (image: %s, calibration: %s)',
                        image.shape, self.intrinsics.size)
            return image

        K = np.array(self.intrinsics.matrix)
        D = np.array(self.intrinsics.distortion)
        h, w = image.shape[:2]
        if self.intrinsics.model == CameraModel.PINHOLE:
            new_K, roi = cv2.getOptimalNewCameraMatrix(K, D, (w, h), 1, (w, h), centerPrincipalPoint=True)

            dst = cv2.undistort(image, K, D, None, new_K)
            if crop:
                x, y, roi_w, roi_h = roi
                dst = dst[y:y+roi_h, x:x+roi_w]
        elif self.intrinsics.model == CameraModel.FISHEYE:
            new_K = self.get_undistorted_camera_matrix(crop=crop)
            map1, map2 = cv2.fisheye.initUndistortRectifyMap(  # pylint: disable=unpacking-non-sequence
                K, D, np.eye(3), new_K, (w, h), cv2.CV_16SC2)
            dst = cv2.remap(image, map1, map2, interpolation=cv2.INTER_LINEAR, borderMode=cv2.BORDER_CONSTANT)
        elif self.intrinsics.model == CameraModel.OMNIDIRECTIONAL:
            flags = cv2.omnidir.RECTIFY_PERSPECTIVE
            new_K = self.get_undistorted_camera_matrix(crop=crop)

            assert self.intrinsics.omnidir_params is not None, 'Omnidirectional parameters are unset'
            R = np.array(self.intrinsics.omnidir_params.rotation.R)
            xi = np.array(self.intrinsics.omnidir_params.xi)
            dst = cv2.omnidir.undistortImage(image, K=K, D=D, xi=xi, Knew=new_K,
                                             flags=flags, new_size=(w, h), R=R)
        else:
            raise ValueError(f'Unknown camera model "{self.intrinsics.model}"')

        return dst

    def get_undistorted_size(self) -> ImageSize:
        """Compute the size of the undistorted image after cropping.

        :return: The size of the undistorted image.
        """
        if self.intrinsics.model == CameraModel.PINHOLE:
            K = np.array(self.intrinsics.matrix)
            D = np.array(self.intrinsics.distortion)
            h, w = self.intrinsics.size.height, self.intrinsics.size.width
            _, roi = cv2.getOptimalNewCameraMatrix(K, D, (w, h), 1, (w, h), centerPrincipalPoint=True)
            return ImageSize(width=roi[2], height=roi[3])
        else:
            return self.intrinsics.size



import abc

from nicegui import Event

from .module import Module, ModuleHardware, ModuleSimulation
from .robot_brain import RobotBrain


class EStop(Module, abc.ABC):
    """A module that detects when the e-stop is triggered.

    The module has a boolean field `active` that is true when the e-stop is triggered.

    There is also a boolean field `is_soft_estop_active` that is true when the soft e-stop is active.
    It can be set to true or false by calling `set_soft_estop(active: bool)`.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self.ESTOP_TRIGGERED = Event[[]]()
        """the e-stop was triggered"""
        self.ESTOP_RELEASED = Event[[]]()
        """the e-stop was released"""
        self._soft_estop_active: bool = False
        self.pressed_estops: list[int] = []

    @property
    def active(self) -> bool:
        """Whether any hardware e-stop or the soft e-stop is active."""
        return any(self.pressed_estops) or self._soft_estop_active

    @property
    def is_soft_estop_active(self) -> bool:
        """Whether the soft e-stop is active."""
        return self._soft_estop_active

    async def set_soft_estop(self, active: bool) -> None:
        """Set the soft e-stop to the given state."""
        self._emit_events(any(self.pressed_estops) or active)
        self._soft_estop_active = active

    def _emit_events(self, value: bool) -> None:
        if value and not self.active:
            self.ESTOP_TRIGGERED.emit()
        if not value and self.active:
            self.ESTOP_RELEASED.emit()


class EStopHardware(EStop, ModuleHardware):
    """Hardware implementation of the e-stop module.

    The module expects a dictionary of pin names and pin numbers.
    """

    def __init__(self, robot_brain: RobotBrain, *, name: str = 'estop', pins: dict[str, int], inverted: bool = True) -> None:
        self.name = name
        self.pins = pins
        self.inverted = inverted
        lizard_code = '\n'.join(f'{name}_{pin} = Input({number})' for pin, number in pins.items())
        if inverted:
            lizard_code += '\n' + '\n'.join(f'{name}_{pin}.inverted = true' for pin in pins)
        core_message_fields = [f'{name}_{pin}.active' for pin in pins]
        super().__init__(robot_brain=robot_brain, lizard_code=lizard_code, core_message_fields=core_message_fields)

    async def set_soft_estop(self, active: bool) -> None:
        await super().set_soft_estop(active)
        await self.robot_brain.send(f'en3.level({"false" if active else "true"})')

    def handle_core_output(self, time: float, words: list[str]) -> None:
        corelist = [words.pop(0) == 'true' for _ in self.pins]
        pressed = [index for index, value in enumerate(corelist) if value]
        if pressed != self.pressed_estops:
            self.log.warning('E-Stop %s changed', pressed)
        self._emit_events(any(pressed) or self.is_soft_estop_active)
        self.pressed_estops[:] = pressed


class EStopSimulation(EStop, ModuleSimulation):
    """Simulation of the e-stop module."""

    async def activate(self) -> None:
        """Activate the soft e-stop."""
        await self.set_soft_estop(True)

    async def deactivate(self) -> None:
        """Deactivate the soft e-stop."""
        await self.set_soft_estop(False)



import io
import logging
from io import BytesIO
from typing import Any

import cv2
import numpy as np
import PIL.Image

from ..geometry import Rectangle
from .image_rotation import ImageRotation
from .turbojpeg_wrapper import TURBO_JPEG

if not TURBO_JPEG:
    logging.getLogger('rosys').warning('TurboJPEG is not available. Using PIL for JPEG decoding and encoding.')


def encode_image_as_jpeg(image: np.ndarray, compression_level: int = 90) -> bytes:
    """Encode image as JPEG using TurboJPEG if available, otherwise PIL."""
    if TURBO_JPEG is not None:
        return TURBO_JPEG.encode(image, quality=compression_level)

    pil_image = PIL.Image.fromarray(image.astype(np.uint8))

    buffer = BytesIO()
    save_kwargs: dict[str, Any] = {}
    if compression_level is not None:
        save_kwargs['quality'] = compression_level

    pil_image.save(buffer, format='JPEG', **save_kwargs)
    return buffer.getvalue()


def decode_jpeg_image(jpeg_bytes: bytes) -> np.ndarray | None:
    """Decode JPEG bytes to NumPy array using TurboJPEG if available, otherwise PIL. Returns None if decoding failed."""
    try:
        if TURBO_JPEG is not None:
            return TURBO_JPEG.decode(jpeg_bytes)
        return np.array(PIL.Image.open(io.BytesIO(jpeg_bytes)))
    except (PIL.UnidentifiedImageError, OSError) as e:
        logging.warning('Failed to decode JPEG image %s', e)
        return None


def process_jpeg_image(data: bytes, rotation: ImageRotation, crop: Rectangle | None = None) -> np.ndarray | None:
    """Rotate and crop a JPEG image and return it as a NumPy pixel array. Returns None if decoding failed."""
    array = decode_jpeg_image(data)
    if array is None:
        return None
    return process_ndarray_image(array, rotation, crop)


def process_ndarray_image(image: np.ndarray, rotation: ImageRotation, crop: Rectangle | None = None) -> np.ndarray:
    """Rotate and crop a NumPy image."""
    if crop is not None:
        image = image[int(crop.y):int(crop.y+crop.height), int(crop.x):int(crop.x+crop.width)]
    if rotation == ImageRotation.LEFT:
        image = cv2.rotate(image, cv2.ROTATE_90_COUNTERCLOCKWISE)
    elif rotation == ImageRotation.RIGHT:
        image = cv2.rotate(image, cv2.ROTATE_90_CLOCKWISE)
    elif rotation == ImageRotation.UPSIDE_DOWN:
        image = cv2.rotate(image, cv2.ROTATE_180)
    return image


def remove_exif(image_data: bytes | bytearray) -> bytes:
    EXIF_MARKER = b'\xFF\xE1'

    pos = 2  # Skip SOI marker
    image_data = bytearray(image_data)
    while pos < len(image_data):
        match_start = image_data.find(EXIF_MARKER, pos)
        if match_start == -1:
            break
        match_end = match_start + 2
        length = int.from_bytes(image_data[match_end:match_end+2], byteorder='big')
        del image_data[match_start:match_end+length]  # Remove EXIF segment
        pos = match_end

    return bytes(image_data)



from dataclasses_json import Exclude, config

from .backup_schedule import BackupSchedule
from .converters import from_dict, replace_dataclass, replace_dict, replace_list, replace_set, to_dict
from .import_export import export_all, export_button, import_all, import_button
from .persistable import Persistable

exclude: dict[str, dict] = config(exclude=Exclude.ALWAYS)

__all__ = [
    'BackupSchedule',
    'Persistable',
    'exclude',
    'export_all',
    'export_button',
    'from_dict',
    'import_all',
    'import_button',
    'replace_dataclass',
    'replace_dict',
    'replace_list',
    'replace_set',
    'to_dict',
]



import asyncio
from datetime import datetime
from typing import Any, Literal

import numpy as np
import socketio
import socketio.exceptions

from .. import rosys
from ..helpers import LazyWorker
from ..persistence import from_dict
from .annotations import Annotations
from .detections import (
    BoxDetection,
    Category,
    ClassificationDetection,
    Detections,
    PointDetection,
    SegmentationDetection,
)
from .detector import Autoupload, Detector, DetectorException, DetectorInfo, ModelVersioningInfo
from .image import Image


class DetectorHardware(Detector):
    """This detector communicates with a zauberzeug detector node via SocketIO.

    It automatically connects and reconnects, submits and receives detections and sends images
    that should be uploaded to the [Zauberzeug Learning Loop](https://zauberzeug.com/products/learning-loop).

    Note: Images must be smaller than ``MAX_IMAGE_SIZE`` bytes (default: 64 MB).
    """
    MAX_IMAGE_SIZE = 64 * 1024 * 1024
    MAX_TIMEOUTS_BEFORE_DISCONNECT = 5

    def __init__(self,
                 *,
                 host: str = 'localhost',
                 port: int = 8004,
                 name: str | None = None,
                 auto_disconnect: bool = True,
                 sio_timeout: int = 5) -> None:
        super().__init__(name=name)

        self.lazy_worker = LazyWorker()
        self.host = host
        self.port = port
        self.auto_disconnect = auto_disconnect

        self.sio_timeout = sio_timeout
        self.timeout_count = 0

        self.sio = socketio.AsyncClient(websocket_extra_options={'max_msg_size': self.MAX_IMAGE_SIZE + 1000})
        self.sio.on('disconnect', lambda: self.log.warning('disconnect on %s:%s', self.host, self.port))
        self.sio.on('connect_error', lambda e: self.log.warning('connect error on %s:%s %s', self.host, self.port, e))

        rosys.on_startup(self.connect)
        rosys.on_repeat(self._ensure_connection, 10.0)

    @property
    def is_connected(self) -> bool:
        return self.sio.connected

    async def _ensure_connection(self) -> None:
        if not self.is_connected:
            self.log.info('trying reconnect %s', self.name)
            if not await self.connect():
                await self.sio.disconnect()
                self.log.error('connection to %s at port %s failed; trying again', self.name, self.port)

    async def connect(self) -> bool:
        try:
            url = f'ws://{self.host}:{self.port}'
            self.log.info('connecting to detector at %s', url)
            await self.sio.connect(url, socketio_path='/ws/socket.io', wait_timeout=3)
            self.log.info('connected successfully')
            return True
        except Exception:
            return False

    async def disconnect(self) -> None:
        await self.sio.disconnect()

    async def upload(self,
                     image: Image,
                     *,
                     tags: list[str] | None = None,
                     source: str | None = None,
                     creation_date: datetime | str | None = None,
                     annotations: Annotations | None = None) -> None:

        if not self.is_connected:
            self.log.error('Upload failed: detector is not connected')
            raise DetectorException('detector is not connected')

        if image.byte_size() > self.MAX_IMAGE_SIZE:
            self.log.error('Upload failed: image too large: %s', image.byte_size())
            raise DetectorException('image too large')

        try:
            self.log.info('Upload detections to port %s', self.port)

            metadata: dict[str, Any] = {
                'source': source,
                'tags': tags or [],
                'creation_date': _creation_date_to_isoformat(creation_date),
            }

            if detections := image.get_detections(self.name):
                detections_dict = detections.to_dict()
                metadata.update({
                    'box_detections': _box_detections_to_int(detections_dict['boxes']),
                    'point_detections': detections_dict['points'],
                    'segmentation_detections': detections_dict['segmentations'],
                    'classification_detections': detections_dict['classifications'],
                })

            if annotations is not None:
                annotations_dict = annotations.to_dict()
                metadata.update({
                    'box_annotations': annotations_dict['box_annotations'],
                    'point_annotations': annotations_dict['point_annotations'],
                    'classification_annotation': annotations_dict['classification_annotation'],
                })

            await self.sio.emit('upload', {
                'image': _array_to_dict(image.array),
                'metadata': metadata,
            })

        except Exception as e:
            self.log.error('Upload failed with exception: %s', e)
            raise DetectorException('Upload failed') from e

    async def detect(self,
                     image: Image,
                     *,
                     lazy: bool = True,
                     autoupload: Autoupload = Autoupload.FILTERED,
                     tags: list[str] | None = None,
                     source: str | None = None,
                     creation_date: datetime | str | None = None) -> Detections | None:

        if not self.is_connected:
            self.log.error('Detection failed: detector is not connected')
            raise DetectorException('detector is not connected')

        if image.byte_size() > self.MAX_IMAGE_SIZE:
            self.log.error('Detection failed: image too large: %s bytes', image.byte_size())
            raise DetectorException('image too large')

        try:
            detect_call = self._detect(image, autoupload, tags or [], source, creation_date)
            if lazy:
                detections = await self.lazy_worker.run(detect_call)
            else:
                detections = await detect_call
            if detections is not None:
                image.set_detections(self.name, detections)
                self.NEW_DETECTIONS.emit(image)
            return detections
        except asyncio.exceptions.CancelledError:
            raise DetectorException('Detection cancelled') from None

    async def _detect(self,
                      image: Image,
                      autoupload: Autoupload,
                      tags: list[str],
                      source: str | None = None,
                      creation_date: datetime | str | None = None) -> Detections:
        try:
            response = await self.sio.call('detect', {
                'image': _array_to_dict(image.array),
                'camera_id': image.camera_id,
                'tags': tags,
                'source': source,
                'autoupload': autoupload.value,
                'creation_date': _creation_date_to_isoformat(creation_date),
            }, timeout=self.sio_timeout)
            self.timeout_count = 0
        except socketio.exceptions.TimeoutError:
            self.timeout_count += 1
            if self.timeout_count > self.MAX_TIMEOUTS_BEFORE_DISCONNECT and self.auto_disconnect:
                self.log.error('Detection timed out %d times in a row. Disconnecting from detector %s',
                               self.timeout_count, self.name)
                await self.disconnect()
            raise DetectorException('Detection timeout') from None
        except Exception as e:
            raise DetectorException('Detection failed') from e

        if not isinstance(response, dict):
            raise DetectorException('Invalid response from detector')
        try:
            detections = _detections_from_dict(response)
        except Exception as e:
            raise DetectorException('Failed to parse detections') from e

        return detections

    async def batch_detect(self,
                           images: list[Image],
                           *,
                           lazy: bool = True,
                           autoupload: Autoupload = Autoupload.FILTERED,
                           tags: list[str] | None = None,
                           source: str | None = None,
                           creation_date: datetime | str | None = None) -> list[Detections] | None:

        if not self.is_connected:
            self.log.error('Detection failed: detector is not connected')
            raise DetectorException('detector is not connected')

        total_size = sum(image.byte_size() for image in images)
        if total_size > self.MAX_IMAGE_SIZE:
            self.log.error('Detection failed: total image size too large: %s bytes', total_size)
            raise DetectorException('total image size too large')

        try:
            detect_call = self._batch_detect(images, autoupload, tags or [], source, creation_date)
            if lazy:
                batch_detections = await self.lazy_worker.run(detect_call)
            else:
                batch_detections = await detect_call

            if batch_detections is not None:
                if len(images) != len(batch_detections):
                    self.log.error('Batch detection failed: number of detections does not match number of images')
                    raise DetectorException('number of detections does not match number of images')

                for image, detections in zip(images, batch_detections, strict=True):
                    image.set_detections(self.name, detections)
                    self.NEW_DETECTIONS.emit(image)
            return batch_detections
        except asyncio.exceptions.CancelledError:
            raise DetectorException('Detection cancelled') from None

    async def _batch_detect(self,
                            images: list[Image],
                            autoupload: Autoupload,
                            tags: list[str],
                            source: str | None = None,
                            creation_date: datetime | str | None = None) -> list[Detections]:

        if len(images) == 0:
            return []

        try:
            # List of image data dictionaries, each with the same structure as the `image` entry in the `detect` endpoint
            images_dict = [_array_to_dict(image.array) for image in images]
            response = await self.sio.call('batch_detect', {
                'images': images_dict,
                'camera_id': images[0].camera_id if images else None,
                'tags': tags,
                'source': source,
                'autoupload': autoupload.value,
                'creation_date': _creation_date_to_isoformat(creation_date),
            }, timeout=self.sio_timeout)
            self.timeout_count = 0
        except socketio.exceptions.TimeoutError:
            self.timeout_count += 1
            if self.timeout_count > self.MAX_TIMEOUTS_BEFORE_DISCONNECT and self.auto_disconnect:
                self.log.error('Detection timed out %d times in a row. Disconnecting from detector %s',
                               self.timeout_count, self.name)
                await self.disconnect()
            raise DetectorException('Detection timeout') from None
        except Exception as e:
            raise DetectorException('Detection failed') from e

        if not isinstance(response, dict):
            raise DetectorException('Invalid response from detector')
        try:
            all_detections = [_detections_from_dict(d) for d in response.get('items', [])]
        except Exception as e:
            raise DetectorException('Failed to parse detections') from e

        return all_detections

    def __str__(self) -> str:
        return f'{type(self).__name__} ({"connected" if self.is_connected else "disconnected"})'

    async def fetch_detector_info(self) -> DetectorInfo:
        if not self.is_connected:
            raise DetectorException('detector is not connected')

        try:
            response = await self.sio.call('about', timeout=self.sio_timeout)
        except socketio.exceptions.TimeoutError:
            self.log.error('Communication timeout for detector %s', self.name)
            raise DetectorException('Communication timeout') from None
        except Exception as e:
            raise DetectorException('Communication failed') from e

        if not isinstance(response, dict):
            raise DetectorException('Invalid response from detector')

        try:
            model_info_dict = response.get('model_info') or {}
            return DetectorInfo(
                operation_mode=response['operation_mode'],
                state=response.get('state'),
                organization=model_info_dict.get('organization'),
                project=model_info_dict.get('project'),
                current_version=model_info_dict.get('version'),
                categories=[
                    Category(uuid=c['id'],
                             name=c['name'],
                             color=c.get('color'),
                             category_type=c.get('type'))
                    for c in model_info_dict.get('categories', [])
                ],
                resolution=model_info_dict.get('resolution'),
                target_version=response.get('target_model'),
                version_control=response['version_control'],
            )
        except Exception as e:
            raise DetectorException('Failed to parse detector info') from e

    async def fetch_model_version_info(self) -> ModelVersioningInfo:
        try:
            response = await self.sio.call('get_model_version', timeout=self.sio_timeout)
        except socketio.exceptions.TimeoutError:
            self.log.error('Communication timeout for detector %s', self.name)
            raise DetectorException('Communication timed out') from None
        except Exception as e:
            raise DetectorException('Communication failed') from e

        if not isinstance(response, dict):
            raise DetectorException('Invalid response from detector')

        try:
            return ModelVersioningInfo(
                current_version=response['current_version'],
                target_version=response['target_version'],
                loop_version=response['loop_version'],
                local_versions=response['local_versions'],
                version_control=response['version_control'],
            )
        except Exception as e:
            raise DetectorException('Failed to parse model versioning info') from e

    async def set_model_version(self, version: Literal['follow_loop', 'pause'] | str) -> None:
        if version not in ['follow_loop', 'pause'] and not version.replace('.', '').isdigit():
            raise DetectorException(f'invalid version control mode: {version} '
                                    f'(allowed: "follow_loop", "pause" or a version number like "1.2")')

        try:
            response = await self.sio.call('set_model_version_mode', version, timeout=self.sio_timeout)
        except socketio.exceptions.TimeoutError:
            self.log.error('Communication timeout for detector %s', self.name)
            raise DetectorException('Communication timeout') from None
        except Exception as e:
            self.log.error('Communication failed for detector %s: %s', self.name, e)
            raise DetectorException('Communication failed') from e

        if not isinstance(response, dict):
            raise DetectorException('Failed to set model version mode')

        if response.get('status') != 'OK':
            error_message = response.get('error', 'unknown error')
            raise DetectorException(f'Failed to set model version mode: {error_message}')

    async def fetch_outbox_mode(self) -> Literal['continuous_upload', 'stopped']:
        try:
            response = await self.sio.call('get_outbox_mode', timeout=self.sio_timeout)
        except socketio.exceptions.TimeoutError:
            self.log.error('Communication timeout for detector %s', self.name)
            raise DetectorException('Communication timeout') from None
        except Exception as e:
            self.log.error('Communication failed for detector %s: %s', self.name, e)
            raise DetectorException('Communication failed') from e

        if not isinstance(response, dict):
            raise DetectorException('Failed to get outbox mode')

        mode = response.get('outbox_mode')
        if mode not in ['continuous_upload', 'stopped']:
            raise DetectorException(f'Invalid outbox mode received: {mode}')

        return mode

    async def set_outbox_mode(self, mode: Literal['continuous_upload', 'stopped']) -> None:
        try:
            response = await self.sio.call('set_outbox_mode', mode, timeout=self.sio_timeout)
        except socketio.exceptions.TimeoutError:
            self.log.error('Communication timeout for detector %s', self.name)
            raise DetectorException('Communication timeout') from None
        except Exception as e:
            self.log.error('Communication failed for detector %s: %s', self.name, e)
            raise DetectorException('Communication failed') from e

        if not isinstance(response, dict):
            raise DetectorException('Failed to set outbox mode')

        if response.get('status') != 'OK':
            error_message = response.get('error', 'unknown error')
            raise DetectorException(f'Failed to set outbox mode: {error_message}')

    async def soft_reload(self) -> None:
        """Trigger a soft reload of the detector.

        :raises DetectorException: if the communication fails.
        """
        try:
            await self.sio.call('soft_reload', timeout=self.sio_timeout)
        except socketio.exceptions.TimeoutError:
            self.log.error('Communication timeout for detector %s', self.name)
            raise DetectorException('Communication timeout') from None
        except Exception as e:
            self.log.error('Communication failed for detector %s: %s', self.name, e)
            raise DetectorException('Communication failed') from e


def _box_detections_to_int(detections: list[dict]) -> list[dict]:
    for detection in detections:
        detection['x'] = int(detection['x'])
        detection['y'] = int(detection['y'])
        detection['width'] = int(detection['width'])
        detection['height'] = int(detection['height'])
    return detections


def _creation_date_to_isoformat(creation_date: datetime | str | None) -> str | None:
    if isinstance(creation_date, str):
        try:
            datetime.fromisoformat(creation_date)  # validate format
        except ValueError as e:
            raise ValueError(f'creation_date string "{creation_date}" is not in valid ISO format') from e
        return creation_date
    elif isinstance(creation_date, datetime):
        return creation_date.isoformat()
    return None


def _detections_from_dict(data: dict[str, Any]) -> Detections:
    return Detections(
        boxes=[from_dict(BoxDetection, d) for d in data.get('box_detections', [])],
        points=[from_dict(PointDetection, d) for d in data.get('point_detections', [])],
        segmentations=[from_dict(SegmentationDetection, d) for d in data.get('segmentation_detections', [])],
        classifications=[from_dict(ClassificationDetection, d) for d in data.get('classification_detections', [])],
    )


def _array_to_dict(array: np.ndarray) -> dict[str, bytes | str | tuple[int, ...]]:
    return {
        'bytes': array.tobytes(order='C'),
        'dtype': str(array.dtype),
        'shape': array.shape,
    }



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, ClassVar, Self

import numpy as np
import PIL.Image
import PIL.ImageDraw

from .. import rosys
from .detections import Detections
from .image_processing import decode_jpeg_image, encode_image_as_jpeg


@dataclass(slots=True, kw_only=True)
class ImageSize:
    width: int
    height: int

    @property
    def tuple(self) -> tuple[int, int]:
        return (self.width, self.height)


ImageArray = np.ndarray[tuple[int, int, int], np.dtype[np.uint8]]
'''An RGB NumPy array with shape (height, width, 3) and dtype np.uint8.'''


@dataclass(slots=True, kw_only=True)
class Image:
    camera_id: str
    time: float  # time of recording
    _array: ImageArray
    _detections: dict[str, Detections] = field(default_factory=dict)
    metadata: dict[str, Any]

    DEFAULT_PLACEHOLDER_SIZE: ClassVar[tuple[int, int]] = (320, 240)

    @property
    def detections(self) -> Detections | None:
        if not self._detections:
            return None
        if len(self._detections) > 1:
            raise RuntimeError(
                f'Image has multiple detection types ({", ".join(self._detections.keys())}). Use `get_detections(type)` instead.')
        return next(iter(self._detections.values()))

    @property
    def array(self) -> ImageArray:
        return self._array

    @property
    def size(self) -> ImageSize:
        return ImageSize(width=self._array.shape[1], height=self._array.shape[0])

    def get_detections(self, detector_id: str) -> Detections | None:
        return self._detections.get(detector_id)

    def set_detections(self, detector_id: str, detections: Detections) -> None:
        self._detections[detector_id] = detections

    @property
    def id(self) -> str:
        return f'{self.camera_id}/{self.time}'

    @classmethod
    def create_placeholder(cls, text: str, *,
                           camera_id: str = 'from_placeholder',
                           time: float | None = None,
                           metadata: dict[str, Any] | None = None,
                           shrink: int = 1) -> Self:
        h, w = cls.DEFAULT_PLACEHOLDER_SIZE
        img = PIL.Image.new('RGB', (h // shrink, w // shrink), color=(73, 109, 137))
        d = PIL.ImageDraw.Draw(img)
        d.text((img.width / 2 - len(text) * 3, img.height / 2 - 5), text, fill=(255, 255, 255))
        return cls.from_array(np.array(img), camera_id=camera_id, time=time, metadata=metadata)

    @classmethod
    def from_pil(cls, pil_image: PIL.Image.Image, *,
                 camera_id: str = 'from_pil',
                 time: float | None = None,
                 metadata: dict[str, Any] | None = None) -> Self:
        """Create an image from a PIL image."""
        return cls.from_array(np.array(pil_image), camera_id=camera_id, time=time, metadata=metadata)

    @classmethod
    def from_jpeg_bytes(cls, jpeg_bytes: bytes, *,
                        camera_id: str = 'from_jpeg_bytes',
                        time: float | None = None,
                        metadata: dict[str, Any] | None = None) -> Self | None:
        """Create an image from plain JPEG bytes. This runs a jpeg decode. Returns None if decoding fails."""
        array = decode_jpeg_image(jpeg_bytes)
        if array is None:
            return None
        if len(array.shape) == 2:
            array = np.repeat(np.expand_dims(array, -1), repeats=3, axis=2)
        return cls.from_array(array, camera_id=camera_id, time=time, metadata=metadata)

    @classmethod
    def from_array(cls, array: np.ndarray, *,
                   camera_id: str = 'from_array',
                   time: float | None = None,
                   metadata: dict[str, Any] | None = None) -> Self:
        """Create an image from a NumPy array."""

        # Convert array to a c-contiguous representation (or do nothing if that
        # is already the case). This ensures fast pickling.
        array = np.ascontiguousarray(array)

        assert array.dtype == np.uint8, 'Array must have dtype np.uint8'
        assert len(array.shape) == 3, 'Array must have shape (height, width, channels)'
        assert array.shape[2] == 3, 'Image should have 3 channels'
        return cls(camera_id=camera_id, time=time or rosys.time(), _array=array, metadata=metadata or {})

    def to_pil(self) -> PIL.Image.Image:
        """Convert the image to a PIL image."""
        return PIL.Image.fromarray(self.array)

    def to_jpeg_bytes(self) -> bytes:
        """Convert the image to a JPEG image by encoding it."""
        return encode_image_as_jpeg(self.array)

    def byte_size(self) -> int:
        """Compute the size of the stored image representation in bytes."""
        return self.array.size



from enum import Enum


class VendorType(Enum):
    JOVISION = 1
    DAHUA = 2
    AXIS = 3
    REOLINK = 4
    UNIARCH = 5
    OTHER = -1


def mac_to_vendor(mac: str) -> VendorType:
    if mac.startswith('e0:62:90'):
        return VendorType.JOVISION
    if mac.startswith(('e4:24:6c', '3c:e3:6b', 'd4:43:0e', 'fc:5f:49', '30:dd:aa')):
        return VendorType.DAHUA
    if mac.startswith('ec:71:db'):
        return VendorType.REOLINK
    if mac.startswith('6c:f1:7e'):
        return VendorType.UNIARCH
    return VendorType.OTHER


def mac_to_url(mac: str, ip: str, substream: int = 0) -> str | None:
    vendor = mac_to_vendor(mac)
    if vendor == VendorType.JOVISION:
        return f'rtsp://admin:admin@{ip}/profile{substream}'
    if vendor == VendorType.DAHUA:
        return f'rtsp://admin:Adminadmin@{ip}/cam/realmonitor?channel=1&subtype={substream}'
    if vendor == VendorType.REOLINK:
        return f'rtsp://admin:Adminadmin@{ip}/Preview_01_{"sub" if substream else "main"}'
    if vendor == VendorType.UNIARCH:
        return f'rtsp://admin:Admin_adm1n@{ip}/media/video{2 if substream else 1}'
    return None



import logging
from datetime import datetime
from pathlib import Path
from uuid import uuid4

import aiofiles
import numpy as np

from ... import run
from ..camera import Camera
from ..image import Image
from .constants import TIME_FORMAT

log = logging.getLogger('rosys.replay_camera')


class ReplayCamera(Camera):

    def __init__(self, *, camera_id: str, images_dir: Path, camera_name: str | None = None) -> None:
        self.images_dir = images_dir
        self.previous_emitted_index = -1
        self.image_paths: list[Path] = []
        self.timestamp_array = np.array([], dtype=float)

        super().__init__(id=camera_id, name=camera_name, connect_after_init=False, base_path_overwrite=str(uuid4()))

        self._load_image_paths(images_dir)

    def _load_image_paths(self, images_dir: Path) -> None:
        """Load images from the specified directory.

        The images should be named with their timestamp using to the format <yyyy-mm-dd_hh-mm-ss.ffffff>.jpg
        """
        path_timestamp_pairs = []
        for file in images_dir.iterdir():
            if file.suffix.lower() not in ['.jpg', '.jpeg', '.png']:
                continue

            try:
                timestamp = datetime.strptime(file.stem, TIME_FORMAT).timestamp()
            except ValueError:
                log.warning('Skipping file %s, invalid timestamp', file)
                continue
            path_timestamp_pairs.append((file, timestamp))

        path_timestamp_pairs.sort(key=lambda x: x[1])
        self.image_paths = [pair[0] for pair in path_timestamp_pairs]
        self.timestamp_array = np.array([pair[1] for pair in path_timestamp_pairs], dtype=float)

    def _find_closest_past_index(self, timestamp: float) -> int:
        return int(np.searchsorted(self.timestamp_array, timestamp, side='right') - 1)

    async def load_image_at_time(self, timestamp: float) -> None:
        """Step to the image closest to the given timestamp."""
        if not self.image_paths:
            return

        if timestamp < self.timestamp_array[0]:
            closest_past_index = 0
        elif timestamp > self.timestamp_array[-1]:
            closest_past_index = len(self.image_paths) - 1
        else:
            closest_past_index = self._find_closest_past_index(timestamp)
        if closest_past_index in [self.previous_emitted_index, -1]:
            return  # No new image to emit

        self.previous_emitted_index = closest_past_index
        image_path = self.image_paths[closest_past_index]

        async with aiofiles.open(image_path, 'rb') as f:
            data = await f.read()

        time = self.timestamp_array[closest_past_index]
        image = await run.cpu_bound(Image.from_jpeg_bytes, data, camera_id=self.id, time=time)
        if image is None:
            return
        self._add_image(image)

    @property
    def is_connected(self) -> bool:
        """To be interpreted as "ready to capture images"."""
        return True



from __future__ import annotations

import abc
import asyncio
import logging
from collections import deque
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from typing import Any, ClassVar, Self

from nicegui import Event

from ... import rosys
from ..image import Image
from ..image_route import create_image_route

logger = logging.getLogger('rosys.vision.camera')


class Camera(abc.ABC):
    IMAGE_HISTORY_LENGTH: ClassVar[int] = 16
    '''How many images are stored in the queue for retrieval via `captured_images` or `get_recent_images`.

    Changing this value only affects cameras created afterwards!
    '''

    def __init__(self,
                 *,
                 id: str,  # pylint: disable=redefined-builtin
                 name: str | None = None,
                 connect_after_init: bool = True,
                 streaming: bool | None = None,
                 polling_interval: float | None = None,
                 base_path_overwrite: str | None = None,
                 **kwargs) -> None:
        super().__init__(**kwargs)
        self.id: str = id
        self.name = name or self.id
        self.connect_after_init = connect_after_init
        self.images: deque[Image] = deque(maxlen=Camera.IMAGE_HISTORY_LENGTH)
        self.base_path: str = f'images/{base_path_overwrite or id}'

        if streaming is not None:
            logger.warning('The `streaming` parameter has been removed. All cameras now stream images by default.')
        if polling_interval is not None:
            logger.warning('The `polling_interval` parameter has been removed. All cameras should now use callbacks.')

        self.NEW_IMAGE = Event[Image]()

        self.device_connection_lock: asyncio.Condition = asyncio.Condition()

        create_image_route(self)

        self._connect_tasks = set()

        if connect_after_init:
            if asyncio.get_event_loop().is_running():
                task = asyncio.create_task(self.connect())
                self._connect_tasks.add(task)
                task.add_done_callback(self._connect_tasks.discard)
            else:
                rosys.on_startup(self.connect)

    @property
    def streaming(self) -> bool:
        logger.warning('The `streaming` parameter has been removed. All cameras now stream images by default.')
        return True

    @streaming.setter
    def streaming(self, value: bool) -> None:  # pylint: disable=unused-argument
        logger.warning('The `streaming` parameter has been removed. All cameras now stream images by default.')

    @property
    def polling_interval(self) -> float:
        logger.warning('The `polling_interval` parameter has been removed. All cameras should now use callbacks.')
        return 0.0

    @polling_interval.setter
    def polling_interval(self, value: float) -> None:  # pylint: disable=unused-argument
        logger.warning('The `polling_interval` parameter has been removed. All cameras should now use callbacks.')

    def get_image_url(self, image: Image) -> str:
        return f'{self.base_path}/{image.time}'

    def get_latest_image_url(self) -> str:
        image = self.latest_captured_image
        if image is None or not self.is_connected:
            return f'{self.base_path}/placeholder'
        return self.get_image_url(image)

    def to_dict(self) -> dict:
        base_path_id = self.base_path.replace('images/', '', 1) if self.base_path.startswith('images') else None
        return {
            'id': self.id,
            'name': self.name,
            'connect_after_init': self.connect_after_init,
            'base_path_overwrite': base_path_id if base_path_id != self.id else None,
        }

    @classmethod
    def args_from_dict(cls, data: dict[str, Any]) -> dict:
        return data

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        return cls(**cls.args_from_dict(data))

    @property
    def is_connected(self) -> bool:
        """To be interpreted as "ready to capture images"."""
        return False

    @asynccontextmanager
    async def _device_connection(self) -> AsyncGenerator[None, None]:
        await self.device_connection_lock.acquire()
        try:
            yield
        finally:
            self.device_connection_lock.release()

    async def connect(self) -> None:  # noqa: B027
        pass

    async def disconnect(self) -> None:  # noqa: B027
        pass

    async def reconnect(self) -> None:
        await self.disconnect()
        await self.connect()

    @property
    def captured_images(self) -> list[Image]:
        return list(self.images)

    @property
    def latest_captured_image(self) -> Image | None:
        return next((i for i in reversed(self.captured_images)), None)

    @property
    def latest_detected_image(self) -> Image | None:
        return next((i for i in reversed(self.captured_images) if i.detections), None)

    def get_recent_images(self, *, current_time: float | None = None, timespan: float = 10.0) -> list[Image]:
        """Returns all images that were captured. Latest images are at the end of the list.

        :param current_time: the starting time for the search; defaults to the current time
        :param timespan: the timespan to search back in seconds
        """
        if current_time is None:
            current_time = rosys.time()
        return [i for i in self.captured_images if i.time > current_time - timespan]

    def _add_image(self, image: Image) -> None:
        self.images.append(image)
        self.NEW_IMAGE.emit(image)

    async def capture_image(self) -> None:
        raise DeprecationWarning('The `capture_image()` method has been removed. All cameras should now use callbacks.')



import asyncio
import logging
from collections.abc import Callable, Coroutine
from typing import Literal

from nicegui import Event

from .. import rosys
from ..driving import Steerer
from .automation import Automation


class Automator:
    """An automator allows running automations, i.e. coroutines that can be paused and resumed.

    See [Click-and-drive](https://rosys.io/examples/click-and-drive/) for a simple example of an automation.

    _steerer_: If provided, manually steering the robot will pause a currently running automation.

    _default_automation_: If provided, it allows the automator to start a new automation without passing an automation
    (e.g. via an "Play"-button like offered by the [automation controls](https://rosys.io/reference/rosys/automation/#rosys.automation.automation_controls)).
    The passed function should return a new coroutine on every call (see [Play-pause-stop](https://rosys.io/examples/play-pause-stop/) example).

    _on_interrupt_: Optional callback that will be called when an automation pauses or stops (the cause is provided as string parameter).

    _notify_: If True, the automator will send notifications when an automation starts, pauses, resumes, stops or fails.
    """

    def __init__(self,
                 steerer: Steerer | None, *,
                 default_automation: Callable | None = None,
                 on_interrupt: Callable | None = None,
                 notify: bool = True) -> None:
        self.AUTOMATION_STARTED = Event[[]]()
        """an automation has been started"""

        self.AUTOMATION_PAUSED = Event[str]()
        """an automation has been paused (string argument: description of the cause)"""

        self.AUTOMATION_RESUMED = Event[[]]()
        """an automation has been resumed"""

        self.AUTOMATION_STOPPED = Event[str]()
        """an automation has been stopped (string argument: description of the cause)"""

        self.AUTOMATION_FAILED = Event[str]()
        """an automation has failed to complete (string argument: description of the cause)"""

        self.AUTOMATION_COMPLETED = Event[[]]()
        """an automation has been completed"""

        self.log = logging.getLogger('rosys.automator')

        self.default_automation = default_automation
        self._on_interrupt = on_interrupt
        self.notify = notify

        self.enabled: bool = True
        self.automation: Automation | None = None

        if steerer:
            steerer.STEERING_STARTED.subscribe(lambda: self.pause(because='steering started'))

        self.AUTOMATION_PAUSED.subscribe(lambda _: self._handle_interrupt())
        self.AUTOMATION_STOPPED.subscribe(lambda _: self._handle_interrupt(stop=True))
        self.AUTOMATION_FAILED.subscribe(lambda _: self._handle_interrupt(stop=True))

        rosys.on_shutdown(lambda: self.stop(because='automator is shutting down'))

    @property
    def is_stopped(self) -> bool:
        return self.automation is None or self.automation.is_stopped

    @property
    def is_stopping(self) -> bool:
        return self.automation is not None and self.automation.is_stopping

    @property
    def is_running(self) -> bool:
        return self.automation is not None and self.automation.is_running

    @property
    def is_paused(self) -> bool:
        return self.automation is not None and self.automation.is_paused

    @property
    def is_pausing(self) -> bool:
        return self.automation is not None and self.automation.is_pausing

    async def _handle_interrupt(self, *, stop: bool = False) -> None:
        assert self.automation is not None
        while self.automation.is_running:
            await rosys.sleep(0.1)
        if self._on_interrupt:
            if asyncio.iscoroutinefunction(self._on_interrupt):
                await self._on_interrupt()
            else:
                self._on_interrupt()
        if stop:
            self.automation = None

    def start(self, coro: Coroutine | None = None, *, paused: bool = False) -> None:
        """Starts a new automation.

        :param coro: the coroutine to start, if None the default automation will be used
        :param paused: whether to start the automation paused
        """
        if coro is None:
            assert self.default_automation is not None
            self.start(self.default_automation(), paused=paused)
            return
        if not self.enabled:
            coro.close()
            return
        self.stop(because='new automation starts')
        self.automation = Automation(coro, self._handle_exception, on_complete=self._on_complete)
        rosys.background_tasks.create(self.automation.run(), name='automation')  # type: ignore
        self.AUTOMATION_STARTED.emit()
        self._notify('automation started')
        if paused:
            self.automation.pause()

    def pause(self, because: str) -> None:
        """Pauses the current automation.

        :param because: the reason for pausing the automation
        """
        if self.is_pausing or self.is_stopping:
            return
        if self.is_running:
            assert self.automation is not None
            self.automation.pause()
            self.AUTOMATION_PAUSED.emit(because)
            self._notify(f'automation paused because {because}')

    def resume(self) -> None:
        """Resumes the current automation."""
        if not self.enabled:
            return
        if self.is_paused:
            assert self.automation is not None
            self.automation.resume()
            self.AUTOMATION_RESUMED.emit()
            self._notify('automation resumed')

    def stop(self, because: str) -> None:
        """Stops the current automation.

        :param because: the reason for stopping the automation
        """
        if self.is_pausing or self.is_stopping:
            return
        if self.is_running or self.is_paused:
            assert self.automation is not None
            self.automation.stop()
            self.AUTOMATION_STOPPED.emit(because)
            self._notify(f'automation stopped because {because}')

    def abort(self, because: str) -> None:
        """Stops the current automation because of a failure.

        :param because: the reason for aborting the automation
        """
        if self.is_stopped:
            return
        assert self.automation is not None
        self.automation.stop()
        self.AUTOMATION_FAILED.emit(because)
        self._notify(f'automation aborted because {because}')

    def enable(self) -> None:
        """Enables the automator.

        It is enabled by default.
        It can be disabled by calling `disable()`.
        """
        self.enabled = True

    def disable(self, because: str) -> None:
        """Disables the automator.

        No automations can be started while the automator is disabled.
        If an automation is running or paused it will be stopped.

        :param because: the reason for disabling the automator
        """
        self.stop(because)
        self.enabled = False

    def set_default_automation(self, default_automation: Callable | None) -> None:
        """Sets the default automation.

        :param default_automation: the default automation to use
        """
        self.default_automation = default_automation

    def _handle_exception(self, e: Exception) -> None:
        self.abort(because=f'an exception occurred in an automation{f": {e}" if str(e) else ""}')
        if rosys.is_test:
            self.log.exception('automation failed')

    def _on_complete(self) -> None:
        self.automation = None
        self.AUTOMATION_COMPLETED.emit()
        self._notify('automation completed', 'positive')

    def _notify(self, message: str,
                type: Literal[  # pylint: disable=redefined-builtin
                    'positive',
                    'negative',
                    'warning',
                    'info',
                    'ongoing',
                ] | None = None) -> None:
        if self.notify:
            rosys.notify(message, type)
        else:
            rosys.log.info(message)



import asyncio
from collections.abc import Coroutine
from enum import StrEnum
from typing import Any, TypeVar

from .. import core

_T = TypeVar('_T')


class SubmissionState(StrEnum):
    CANCELED = 'canceled'
    WAITING = 'waiting'
    SCHEDULED = 'scheduled'


class SubmissionData:
    def __init__(self, state: SubmissionState):
        self.state = state
        self.can_progress: asyncio.Event = asyncio.Event()


class QueueData:
    def __init__(self):
        self.last_run_at: float = -float('inf')
        self.waiting_request: SubmissionData | None = None


class MultiQueueLazyWorker:
    """A scheduling worker that controls access to an (implicit) shared resource.

    The shared resource is not managed by the worker itself, but assumed
    to be accessed by the coroutine given to `run`. The maximum number of
    concurrent accesses is controlled by the `running_slots` parameter.

    Items from the queues are scheduled fairly, favouring the queue with the
    least recent access to the resource. If the resource is equally contended
    by all queues, this results in round-robin scheduling.

    Each queue has a fixed size of one, where new requests (by calls to `run`)
    displace old ones, similar to the mechanism used by the `LazyWorker`.

    *Note*: The implementation is *not* threadsafe, so it should not be used to
    synchronize access from different threads (e.g., by using a multi threaded
    async runtime).
    """

    def __init__(self, running_slots: int = 1):
        """ Initialize the worker with a fixed number of running slots.

        :param running_slots: Maximum number coroutines running concurrently.
        """

        self.queues: dict[str, QueueData] = {}
        self.running_slots_available: int = running_slots

    async def run(self, queue_id: str, coro: Coroutine[Any, Any, _T]) -> _T | None:
        """Try to run the given coroutine. A later call with the same
        queue_id` will displace this one. In that case `None` is returned.

        :param queue_id: Uniquely identifies the queue that coro will be pushed to
        :param coro: Coroutine which accesses a shared resource
        :return: None if a later call displaced this one, the return value of
                 the coroutine otherwise
        """

        data = SubmissionData(state=SubmissionState.WAITING)

        # Make sure we have a client for our id
        if queue_id not in self.queues:
            self.queues[queue_id] = QueueData()
        client = self.queues[queue_id]

        # Displace waiting request by the current (more recent) one
        if client.waiting_request is not None:
            displaced = client.waiting_request
            displaced.state = SubmissionState.CANCELED
            displaced.can_progress.set()
        client.waiting_request = data

        # If there are running slots available, take one immediately
        if self.running_slots_available > 0:
            self.running_slots_available -= 1
            data.state = SubmissionState.SCHEDULED
            client.waiting_request = None

        while True:
            if data.state == SubmissionState.CANCELED:
                coro.close()
                return None

            if data.state == SubmissionState.SCHEDULED:
                client.last_run_at = core.time()
                try:
                    return await coro
                finally:
                    next_client = min(((k, v) for k, v in self.queues.items() if v.waiting_request is not None),
                                      key=lambda kv: kv[1].last_run_at,
                                      default=None, )

                    if next_client is None:
                        # Make running slot available
                        self.running_slots_available += 1
                    else:
                        # Pass this running slot to next request
                        next_client_id, _ = next_client
                        next_request = self.queues[next_client_id].waiting_request
                        assert next_request is not None
                        next_request.state = SubmissionState.SCHEDULED
                        self.queues[next_client_id].waiting_request = None
                        next_request.can_progress.set()

            await data.can_progress.wait()
            data.can_progress.clear()



from __future__ import annotations

from dataclasses import asdict, dataclass, field


@dataclass(slots=True, kw_only=True)
class BoxAnnotation:
    category_name: str
    x: int
    y: int
    width: int
    height: int


@dataclass(slots=True, kw_only=True)
class PointAnnotation:
    category_name: str
    x: float
    y: float


@dataclass(slots=True, kw_only=True)
class ClassificationAnnotation:
    category_name: str


@dataclass(slots=True, kw_only=True)
class Annotations:
    box_annotations: list[BoxAnnotation] = field(default_factory=list)
    point_annotations: list[PointAnnotation] = field(default_factory=list)
    classification_annotation: ClassificationAnnotation | None = None

    def to_dict(self) -> dict:
        return asdict(self)



from __future__ import annotations

import asyncio
import logging
import re
import shlex
import signal
import struct
import subprocess
from asyncio.subprocess import Process
from collections.abc import AsyncGenerator, Awaitable, Callable
from dataclasses import dataclass
from enum import Enum
from typing import Literal

import numpy as np
from nicegui import background_tasks

from ... import rosys
from ...vision.image import ImageArray
from .jovision_rtsp_interface import JovisionInterface
from .vendors import VendorType, mac_to_url, mac_to_vendor


class RtspDevice:

    def __init__(self, mac: str, ip: str, *,
                 substream: int, fps: int, on_new_image_data: Callable[[ImageArray, float], Awaitable | None],
                 avdec: Literal['h264', 'h265'] = 'h264') -> None:
        self._mac = mac
        self._ip = ip
        self.log = logging.getLogger('rosys.vision.rtsp_camera.rtsp_device.' + self._mac)

        self._fps = fps
        self._substream = substream
        self._on_new_image_data = on_new_image_data
        self._avdec = avdec

        self._capture_task: asyncio.Task | None = None
        self._capture_process: Process | None = None
        self._authorized: bool = True

        vendor_type = mac_to_vendor(mac)

        self._settings_interface: JovisionInterface | None = None
        if vendor_type == VendorType.JOVISION:
            self._settings_interface = JovisionInterface(ip)
        else:
            self.log.warning('[%s] No settings interface for vendor type %s', self._mac, vendor_type)
            self.log.warning('[%s] Using default fps of 10', self._mac)

        self.log.info('[%s] Starting VideoStream for %s', self._mac, self.url)
        self._start_gstreamer_task()

    @property
    def is_connected(self) -> bool:
        return self._capture_task is not None

    @property
    def authorized(self) -> bool:
        return self._authorized

    @property
    def url(self) -> str:
        url = mac_to_url(self._mac, self._ip, self._substream)
        if url is None:
            raise ValueError(f'could not determine RTSP URL for {self._mac}')
        return url

    async def shutdown(self) -> None:
        if self._capture_process is not None:
            self.log.debug('[%s] Terminating gstreamer process', self._mac)
            self._capture_process.terminate()
            try:
                await asyncio.wait_for(self._capture_process.wait(), timeout=5)
            except TimeoutError:
                self.log.warning('[%s] Timeout while waiting for gstreamer process to terminate', self._mac)
            else:
                self.log.debug('[%s] Successfully shut down process (code %s)',
                               self._mac, self._capture_process.returncode if self._capture_process.returncode is not None else 'None')
                self._capture_process = None
        if self._capture_task is not None and not self._capture_task.done():
            self.log.debug('[%s] Cancelling gstreamer task', self._mac)
            self._capture_task.cancel()
            try:
                await asyncio.wait_for(self._capture_task, timeout=5)
            except TimeoutError:
                self.log.warning('[%s] Timeout while waiting for capture task to cancel', self._mac)
                return
            except asyncio.CancelledError:
                self.log.debug('[%s] Task was successfully cancelled', self._mac)
            else:
                self.log.debug('[%s] Task finished', self._mac)
            self._capture_task = None

    def _start_gstreamer_task(self) -> None:
        self.log.debug('[%s] Starting gstreamer task', self._mac)
        if self._capture_task is not None and not self._capture_task.done():
            self.log.warning('[%s] capture task already running', self._mac)
            return
        self._capture_task = background_tasks.create(self._run_gstreamer(), name=f'capture {self._mac}')

    async def restart_gstreamer(self) -> None:
        await self.shutdown()
        self._start_gstreamer_task()

    async def _run_gstreamer(self) -> None:
        if self._capture_process is not None and self._capture_process.returncode is None:
            self.log.warning('[%s] capture process already running', self._mac)
            return

        async def stream() -> AsyncGenerator[ImageArray, None]:
            url = self.url
            self.log.debug('[%s] Starting gstreamer pipeline for %s', self._mac, url)
            # to try: replace avdec_h264 with nvh264dec ! nvvidconv (!videoconvert)
            command = f'gst-launch-1.0 --quiet rtspsrc location="{url}" latency=0 protocols=tcp ! rtp{self._avdec}depay ! avdec_{self._avdec} ! videoconvert ! video/x-raw,format=RGB ! queue max-size-buffers=1 leaky=downstream ! gdppay ! fdsink sync=false'
            self.log.debug('[%s] Running command: %s', self._mac, command)
            process = await asyncio.create_subprocess_exec(
                *shlex.split(command),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                limit=1024*1024*1024,
            )
            assert process.stdout is not None
            assert process.stderr is not None
            self._capture_process = process

            width = None
            height = None
            while process.returncode is None:
                assert process.stdout is not None

                try:
                    packet = await GDPPacket.read(process.stdout)
                except asyncio.exceptions.IncompleteReadError:
                    break

                if packet.payload_type == GDPPayloadType.CAPS:
                    cap_text = packet.payload.decode('utf-8', 'ignore')

                    w = GDP_CAPS_WIDTH_REGEX.search(cap_text)
                    h = GDP_CAPS_HEIGHT_REGEX.search(cap_text)

                    assert w is not None and h is not None
                    assert len(w.groups()) == 1
                    assert len(h.groups()) == 1

                    width = int(w.group(1))
                    height = int(h.group(1))

                elif packet.payload_type == GDPPayloadType.BUFFER:
                    assert width is not None and height is not None

                    assert width * height * 3 == len(packet.payload)
                    frame = np.frombuffer(packet.payload, dtype=np.uint8).reshape(height, width, 3)

                    yield frame

            try:
                await asyncio.wait_for(process.wait(), timeout=5)
            except TimeoutError:
                self.log.warning(
                    '[%s] Stream ended. Timeout while waiting for gstreamer process to terminate', self._mac)
                return

            return_code = process.returncode
            if return_code == -1 * signal.SIGTERM:
                self.log.debug('gstreamer process %s was terminated using SIGTERM', process.pid)
            else:
                error = await process.stderr.read()
                error_message = error.decode()
                self.log.error('gstreamer process %s exited with code %s.\nstderr: %s',
                               process.pid, return_code, error_message)

                if 'Unauthorized' in error_message:
                    self._authorized = False

        async for image in stream():
            timestamp = rosys.time()
            result = self._on_new_image_data(image, timestamp)
            if isinstance(result, Awaitable):
                await result

        self.log.info('[%s] stream ended', self._mac)

        self._capture_task = None

    async def set_fps(self, fps: int) -> None:
        self._fps = fps

        if self._settings_interface is not None:
            await self._settings_interface.set_fps(stream_id=self._substream, fps=self._fps)

    async def get_fps(self) -> int | None:
        if self._settings_interface is not None:
            return await self._settings_interface.get_fps(stream_id=self._substream)
        return self._fps

    def set_substream(self, index: int) -> None:
        self._substream = index

    def get_substream(self) -> int:
        return self._substream

    async def set_bitrate(self, bitrate: int) -> None:
        if self._settings_interface is not None:
            await self._settings_interface.set_bitrate(stream_id=self._substream, bitrate=bitrate)

    async def get_bitrate(self) -> int | None:
        if self._settings_interface is not None:
            return await self._settings_interface.get_bitrate(stream_id=self._substream)
        return None

    def get_avdec(self) -> Literal['h264', 'h265'] | None:
        return self._avdec

    def set_avdec(self, avdec: Literal['h264', 'h265']) -> None:
        self._avdec = avdec


class GDPPayloadType(Enum):
    NONE = 0
    BUFFER = 1
    CAPS = 2
    EVENT_NONE = 3


# See https://maemo.org/api_refs/5.0/5.0-final/gstreamer-libs-0.10/gstreamer-libs-gstdataprotocol.html for header format
GDPPACKET_FORMAT = struct.Struct('>HcxHIQQQQH14sHH')
GDP_CAPS_WIDTH_REGEX = re.compile(r'width=\(int\)\s*(\d+)')
GDP_CAPS_HEIGHT_REGEX = re.compile(r'height=\(int\)\s*(\d+)')
GDP_HEADER_SIZE = 62


@dataclass(slots=True, kw_only=True)
class GDPPacket:
    payload_type: GDPPayloadType
    payload: bytes

    @staticmethod
    async def read(stream: asyncio.StreamReader) -> GDPPacket:
        header_bytes = await stream.readexactly(GDP_HEADER_SIZE)
        _version, _flags, gdp_type, length, *_ = GDPPACKET_FORMAT.unpack(header_bytes)
        return GDPPacket(
            payload_type=GDPPayloadType(gdp_type) if gdp_type < 3 else GDPPayloadType.EVENT_NONE,
            payload=await stream.readexactly(length),
        )



from .image_recorder import ImageRecorder
from .replay_camera import ReplayCamera
from .replay_camera_provider import ReplayCameraProvider

__all__ = [
    'ImageRecorder',
    'ReplayCamera',
    'ReplayCameraProvider',
]



import logging
from datetime import datetime
from pathlib import Path

import aiofiles

from ... import run
from ..camera_provider import CameraProvider
from ..image import Image
from ..image_processing import encode_image_as_jpeg
from .constants import TIME_FORMAT

log = logging.getLogger('rosys.image_recorder')


class ImageRecorder:
    """This module saves images from a CameraProvider to disk."""

    def __init__(self, camera_provider: CameraProvider, data_dir: Path) -> None:
        """This module saves images from a CameraProvider to disk.

        :param camera_provider: the camera provider to record images from
        :param data_dir: the directory to save images to. Images are saved in subdirectories
        """
        self.camera_provider = camera_provider
        self.data_dir = data_dir.expanduser()
        self._recording = False

    @property
    def recording(self) -> bool:
        return self._recording

    def set_recording(self, value: bool):
        """Enable or disable recording images from the camera provider."""
        self._recording = value
        if self._recording:
            self.camera_provider.NEW_IMAGE.subscribe(self._save_image)
        else:
            self.camera_provider.NEW_IMAGE.unsubscribe(self._save_image)

    async def _save_image(self, image: Image):
        """Save the provided image to the data directory.

        Images are saved to a subfolder named after the camera ID (with ':' replaced by '--'),
        using filenames in the format YYYY-MM-DD_HH-MM-SS.ffffff.jpg based on the image timestamp.

        :param image: the image to save
        """
        path = self.data_dir / image.camera_id.replace(':', '--')
        path.mkdir(parents=True, exist_ok=True)

        file_path = path / f'{datetime.fromtimestamp(image.time).strftime(TIME_FORMAT)}.jpg'

        async with aiofiles.open(file_path, 'wb') as f:
            jpeg_bytes = await run.cpu_bound(encode_image_as_jpeg, image.array)
            if jpeg_bytes is None:
                return  # NOTE: This only happens when stopping
            await f.write(jpeg_bytes)
            log.debug('Saved image from camera %s to %s', image.camera_id, file_path)



from .gnss import Gnss, GnssMeasurement
from .gnss_hardware import GnssHardware
from .gnss_simulation import GnssSimulation
from .nmea import Gga, GpsQuality, Gst, Pssn

__all__ = [
    'Gga',
    'Gnss',
    'GnssHardware',
    'GnssMeasurement',
    'GnssSimulation',
    'GpsQuality',
    'Gst',
    'Pssn',
]



from __future__ import annotations

import asyncio
import functools
import logging
from collections.abc import Callable, Coroutine, Generator
from contextvars import ContextVar
from typing import Any

# context for the currently running Automation; used by @uninterruptible
_CURRENT_AUTOMATION: ContextVar[Automation | None] = ContextVar('rosys_automation_current', default=None)


def uninterruptible(func: Callable):
    """Decorator to make an async function uninterruptible until it exits.

    Note that ``rosys.automation.parallelize`` will also be uninterruptible if one of its coroutines is marked with this decorator.
    """
    @functools.wraps(func)
    async def _wrapped(*args, **kwargs):
        automation = _CURRENT_AUTOMATION.get()
        try:
            if automation is not None:
                automation._uninterruptible_depth += 1
            result = await func(*args, **kwargs)
        finally:
            if automation is not None:
                automation._uninterruptible_depth -= 1
                await asyncio.sleep(0)
        return result
    return _wrapped


class Automation:
    """An automation wraps a coroutine and allows pausing and resuming it.

    Optional exception and completion handlers are called if provided.
    This class is used internally by the [automator](https://rosys.io/reference/rosys/automation/#rosys.automation.Automator).
    """

    def __init__(self,
                 coro: Coroutine,
                 exception_handler: Callable | None = None,
                 on_complete: Callable | None = None) -> None:
        self.log = logging.getLogger('rosys.automation')
        self.coro = coro
        self.exception_handler = exception_handler
        self.on_complete = on_complete
        self._can_run = asyncio.Event()
        self._can_run.set()
        self._stop = False
        self._is_waited = False
        self._uninterruptible_depth = 0  # >0 while inside an @uninterruptible section

    @property
    def is_running(self) -> bool:
        return not self.is_stopped and not self.is_paused

    @property
    def is_paused(self) -> bool:
        return self._is_waited and not self._can_run.is_set() and self._uninterruptible_depth == 0

    @property
    def is_pausing(self) -> bool:
        return self._is_waited and not self._can_run.is_set() and self._uninterruptible_depth > 0

    @property
    def is_stopped(self) -> bool:
        return not self._is_waited

    @property
    def is_stopping(self) -> bool:
        return self._stop and not self.is_stopped

    async def run(self) -> Any | None:
        return await self

    def __await__(self) -> Generator[Any, None, Any | None]:
        token = _CURRENT_AUTOMATION.set(self)  # bind this Automation instance into the task context
        try:
            self._is_waited = True
            coro_iter = self.coro.__await__()
            iter_send, iter_throw = coro_iter.send, coro_iter.throw
            send: Callable = iter_send
            message: Any = None
            while not (self._uninterruptible_depth == 0 and self._stop):
                try:
                    while self._uninterruptible_depth == 0 and not self._can_run.is_set() and not self._stop:
                        yield from self._can_run.wait().__await__()  # pylint: disable=no-member
                except BaseException as err:
                    send, message = iter_throw, err

                if self._uninterruptible_depth == 0 and self._stop:
                    return None

                try:
                    signal = send(message)
                except StopIteration as err:
                    self.log.info('automation is finished')
                    if self.on_complete:
                        self.on_complete()
                    return err.value
                send = iter_send
                try:
                    message = yield signal
                except BaseException as err:
                    send, message = iter_throw, err
        except Exception as e:
            self.log.exception('automation failed')
            if self.exception_handler:
                self.exception_handler(e)
            raise
        finally:
            self._is_waited = False
            _CURRENT_AUTOMATION.reset(token)
        return None

    def pause(self) -> None:
        self._can_run.clear()

    def resume(self) -> None:
        self._can_run.set()

    def stop(self) -> None:
        self._can_run.set()
        self._stop = True



import json
from collections.abc import Callable
from pathlib import Path
from typing import Any

from nicegui import events, ui

from .. import helpers
from .persistable import Persistable


def export_all(path: Path | None = None, *, indent: int = 4) -> dict[str, dict[str, Any]]:
    """Export all persistable objects to a dictionary or file."""
    data = {key: instance.backup_to_dict() for key, instance in Persistable.instances.items()}
    if path:
        path.write_text(json.dumps(data, indent=indent))
    return data


def import_all(source: Path | dict[str, dict[str, Any]]) -> None:
    """Import all persistable objects from a file or dictionary.

    Note that objects that are not in the source dictionary are silently skipped.
    Exceptions that occur while restoring an individual object are not caught.
    """
    if isinstance(source, Path):
        data = json.loads(source.read_text())
    else:
        data = source
    for key, instance in Persistable.instances.items():
        if key in data:
            instance.restore_from_dict(data[key])


def export_button(title: str = 'Export', *,
                  icon: str | None = None,
                  color: str | None = 'primary',
                  indent: int = 4) -> ui.button:
    return ui.button(title, icon=icon, color=color,
                     on_click=lambda: ui.download.content(json.dumps(export_all(), indent=indent),
                                                          filename='export.json'))


def import_button(title: str = 'Import', *,
                  icon: str | None = None,
                  color: str | None = 'primary',
                  on_completion: Callable | None = None) -> ui.button:
    async def restore_from_file(e: events.UploadEventArguments) -> None:
        import_all(await e.file.json())
        dialog.close()
        if on_completion is not None:
            await helpers.invoke(on_completion)
    with ui.dialog() as dialog:
        ui.upload(label=title, on_upload=restore_from_file).props('flat')
    return ui.button(title, icon=icon, color=color, on_click=dialog.open)
